# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path
from os import listdir
import re
from glob import glob

import numpy as np
from scipy.linalg import inv, solve_triangular, svd, pascal, invpascal
from scipy.optimize import nnls

import abel
from abel.tools.vmi import Distributions
from abel.tools.symmetry import put_image_quadrants

###############################################################################
#
# rBasex — Abel transform for velocity-map images (~spherical harmonics),
#          similar to pBasex, but with analytical basis functions and without
#          pixel-grid transformation; a simplified version of the method
#          described in M. Ryazanov Ph.D. dissertation.
#
# pBasex:
#   Gustavo A. Garcia, Laurent Nahon, Ivan Powis,
#   “Two-dimensional charged particle image inversion using a polar basis
#    function expansion”,
#   Review of Scientific Instruments 75, 4989 (2004).
#   http://dx.doi.org/10.1063/1.1807578
#
# Ryazanov:
#   Mikhail Ryazanov,
#   “Development and implementation of methods for sliced velocity map imaging.
#    Studies of overtone-induced dissociation and isomerization dynamics of
#    hydroxymethyl radical (CH₂OH and CD₂OH)”,
#   Ph.D. dissertation, University of Southern California, 2012.
#   https://search.proquest.com/docview/1289069738
#   https://digitallibrary.usc.edu/asset-management/2A3BF169XWB4
#
###############################################################################


# Caches and their parameters
_prm = None  # [shape, origin, rmax, order, odd]
_weights = None  # weights — pixel weights
_dst = None  # Distributions object
_bs_prm = None  # [Rmax, order, odd]
_bs = None  # [P[n]] — projected functions
_ibs = None  # [rbin, wl, wu, cos^n] — arrays for image construction
_trf = None  # [Af[n]] — forward transform matrices
_tri_full = None  # [Ai[n]] — inverse-transform matrices without mask and reg
_tri_prm = None  # [reg] — regularization parameters
_tri = None  # [Ai[n]] — inverse-transform matrices (or Af for reg='pos')


def rbasex_transform(IM, origin='center', rmax='MIN', order=2, odd=False,
                     weights=None, direction='inverse', reg=None, out='same',
                     basis_dir=None, verbose=False):
    r"""
    This function takes the input image and outputs its forward or inverse Abel
    transform as an image and its radial distributions.

    The **origin**, **rmax**, **order**, **odd** and **weights** parameters are
    passed to :class:`abel.tools.vmi.Distributions`, so see its documentation
    for their detailed descriptions.

    Parameters
    ----------
    IM : m × n numpy array
        the image to be transformed
    origin : tuple of int or str
        image origin, explicit in the (row, column) format, or as a location
        string (by default, the image center)
    rmax : int or string
        largest radius to include in the transform (by default, the largest
        radius with at least one full quadrant of data)
    order : int
        highest angular order present in the data, ≥ 0 (by default, 2). Working
        with very high orders (≳ 15) can result in excessive noise, especially
        at small radii and for narrow peaks.
    odd : bool
        include odd angular orders (by default is `False`, but is enabled
        automatically if **order** is odd)
    weights : m × n numpy array, optional
        weighting factors for each pixel. The array shape must match the image
        shape. Parts of the image can be excluded from analysis by assigning
        zero weights to their pixels. By default is `None`, which applies equal
        weight to all pixels.
    direction : str: ``'forward'`` or ``'inverse'``
        type of Abel transform to be performed (by default, inverse)
    reg : None or str or tuple (str, float), optional
        regularization to use for inverse Abel transform. ``None`` (default)
        means no regularization, a string selects a non-parameterized
        regularization method, and parameterized methods are selected by a
        tuple (`method`, `strength`). Available methods are:

        ``('L2', strength)``:
            Tikhonov :math:`L_2` regularization with `strength` as the square
            of the Tikhonov factor. This is the same as “Tikhonov
            regularization” used in BASEX, with almost identical effects on the
            radial distributions.
        ``('diff', strength)``:
            Tikhonov regularization with the difference operator (approximation
            of the derivative) multiplied by the square root of `strength` as
            the Tikhonov matrix. This tends to produce less blurring, but more
            negative overshoots than ``'L2'``.
        ``('SVD', strength)``:
            truncated SVD (singular value decomposition) with
            N = `strength` × **rmax** largest singular values removed for each
            angular order. This mimics the approach proposed (but in fact not
            used) in pBasex. `Not recommended` due to generally poor results.
        ``'pos'``:
            non-parameterized method, finds the best (in the least-squares
            sense) solution with non-negative :math:`\cos^n\theta \sin^m\theta`
            terms (see :meth:`~abel.tools.vmi.Distributions.Results.cossin`).
            For **order** = 0, 1, and 2 (with **odd** = `False`) this is
            equivalent to :math:`I(r, \theta) \geqslant 0`; for higher orders
            this assumption is stronger than :math:`I \geqslant 0` and
            corresponds to no interference between different multiphoton
            channels. Not implemented for odd orders > 1.

            Notice that this method is nonlinear, which also means that it is
            considerably slower than the linear methods and the transform
            operator cannot be cached.

        In all cases, `strength` = 0 provides no regularization. For the
        Tikhonov methods, `strength` ~ 100 is a reasonable value for megapixel
        images. For truncated SVD, `strength` must be < 1; `strength` ~ 0.1 is
        a reasonable value; `strength` ~ 0.5 can produce noticeable ringing
        artifacts. See the :ref:`full description <rBasexmathreg>` and examples
        there.
    out : str or None
        shape of the output image:

        ``'same'`` (default):
            same shape and origin as the input
        ``'fold'`` (fastest):
            Q0 (upper right) quadrant (for ``odd=False``) or right half (for
            ``odd=True``) up to **rmax**, but limited to the largest
            input-image quadrant (or half)
        ``'unfold'``:
            like ``'fold'``, but symmetrically “unfolded” to all 4 quadrants
        ``'full'``:
            all pixels with radii up to **rmax**
        ``'full-unique'``:
            the unique part of ``'full'``: Q0 (upper right) quadrant for
            ``odd=False``, right half for ``odd=True``
        ``None``:
            no image (**recon** will be ``None``). Can be useful to avoid
            unnecessary calculations when only the transformed radial
            distributions (**distr**) are needed.
    basis_dir : str, optional
        path to the directory for saving / loading the basis set (useful only
        for the inverse transform without regularization; time savings in other
        cases are small and might be negated by the disk-access overhead). Use
        ``''`` for the default directory. If ``None`` (default), the basis set
        will not be loaded from or saved to disk.
    verbose : bool
        print information about processing (for debugging), disabled by default

    Returns
    -------
    recon : 2D numpy array or None
        the transformed image. Is centered and might have different dimensions
        than the input image.
    distr : Distributions.Results object
        the object from which various distributions for the transformed image
        can be retrieved, see :class:`abel.tools.vmi.Distributions.Results`
    """
    if order == 0:
        odd = False  # (to eliminate additional checks)
    elif order % 2:
        odd = True  # enable automatically for odd orders

    # extract radial profiles from input image
    p = _profiles(IM, origin, rmax, order, odd, weights, verbose)
    # (caches Distributions as _dst)

    Rmax = _dst.rmax

    # get appropriate transform matrices
    A = get_bs_cached(Rmax, order, odd, direction, reg, _dst.valid,
                      basis_dir, verbose)

    # transform radial profiles
    if reg == 'pos':
        if verbose:
            print('Solving NNLS equations...')
        N = len(p)
        p = np.hstack(p)
        cs = nnls(A, p)[0]
        cs = np.split(cs, N)
        if odd:
            # (1 ± cos) / 2 → cos^0, cos^1
            c = [cs[0] + cs[1], cs[0] - cs[1]]
        else:
            # cossin → cos transform
            C = np.flip(invpascal(N, 'upper'))
            c = C.dot(cs)
    else:
        if verbose:
            print('Applying radial transforms...')
        c = [An.dot(pn) for An, pn in zip(A, p)]

    # construct output (transformed) distributions
    distr = Distributions.Results(np.arange(Rmax + 1), np.array(c),
                                  order, odd,
                                  _dst.valid)

    if out is None:
        return None, distr

    # output size
    if out == 'same':
        height = _dst.shape[0] if odd else _dst.VER + 1
        width = _dst.HOR + 1
        row = _dst.row if odd else 0
    elif out in ['fold', 'unfold']:
        height = _dst.Qheight
        width = _dst.Qwidth
        row = _dst.row if odd else 0
    elif out in ['full', 'full-unique']:
        height = 2 * Rmax + 1 if odd else Rmax + 1
        width = Rmax + 1
        row = Rmax if odd else 0
    else:
        raise ValueError('Wrong output shape "{}"'.format(out))
    # construct output image from transformed radial profiles
    if verbose:
        print('Constructing output image...')
    # bottom right quadrant or right half
    recon = _image(height, width, row, c, verbose)
    if odd:
        if out not in ['fold', 'full-unique']:
            # combine with left half (mirrored without central column)
            recon = np.hstack((recon[:, :0:-1], recon))
    else:  # even only
        recon = recon[::-1]  # flip to Q0
        if out not in ['fold', 'full-unique']:
            # assemble full image
            recon = put_image_quadrants((recon, recon, recon, recon),
                                        (2 * height - 1, 2 * width - 1))
    if out == 'same':
        # crop as needed
        row = 0 if odd else _dst.VER - _dst.row
        col = _dst.HOR - _dst.col
        H, W = IM.shape
        recon = recon[row:row + H, col:col + W]

    return recon, distr


def _profiles(IM, origin, rmax, order, odd, weights, verbose):
    """
    Get radial profiles of cos^n theta terms from the input image.
    """
    # the Distributions object is cached to speed up further calculations,
    # plus its cos^n theta matrices are used later to construct the transformed
    # image
    global _prm, _weights, _dst, _ibs, _trf, _tri_prm, _tri

    old_valid = None if _dst is None else _dst.valid

    if verbose:
        print('Extracting radial profiles...')
    prm = [IM.shape, origin, rmax, order, odd]
    if _prm != prm or _weights is not weights:
        _prm = prm
        _weights = weights
        _dst = Distributions(origin=origin, rmax=rmax, order=order, odd=odd,
                             weights=weights, use_sin=False, method='linear')
        if verbose:
            print('(new Distributions object created)')
        # reset image basis
        _ibs = None
    else:
        if verbose:
            print('(reusing cached Distributions object)')

    c = _dst(IM).cos()

    if not np.array_equal(_dst.valid, old_valid):
        # reset transforms
        _trf = None
        _tri_prm = None
        _tri = None

    return c


def _get_image_bs(height, width, row, verbose):
    global _ibs

    if _ibs is not None:
        if verbose:
            print('(using cached image basis)')
        return _ibs

    # _dst quadrant has the minimal size, so height and width either equal its
    # dimensions, or at least one of them is larger
    if height == _dst.Qheight and width == _dst.Qwidth:
        if verbose:
            print('(using image basis from Distributions object)')
        # use arrays already computed in _dst
        _ibs = [_dst.bin, _dst.wl, _dst.wu, _dst.c]
    else:  # height > _dst.Qheight or width > _dst.Qwidth
        # compute arrays of requested size
        rmax = _dst.rmax
        # x row
        x = np.arange(float(width))
        # y and y^2 columns
        y = row - np.arange(float(height))[:, None]
        y2 = y**2
        # arrays of r^2 and r
        r2 = x**2 + y2
        r = np.sqrt(r2)
        # radial bins (as "indexing integers")
        rbin = r.astype(np.intp)  # round down (floor)
        rbin[rbin > rmax] = rmax + 1  # last bin is then discarded
        # weights for upper and lower bins
        wu = r - rbin
        wl = 1 - wu
        # cos^n theta
        cos = [None]  # (cos^0 theta is not used)
        if _dst.odd:
            r[row, 0] = np.inf  # (avoid division by zero)
            cos.append(y / r)  # cos^1 theta
        else:
            r2[0, 0] = np.inf  # (avoid division by zero; row = 0)
            cos.append(y2 / r2)  # cos^2 theta
        for n in range(2, len(_dst.c)):  # remaining powers
            cos.append(cos[1] * cos[n - 1])
        if verbose:
            print('(image basis constructed)')

        _ibs = [rbin, wl, wu, cos]

    return _ibs


def _image(height, width, row, c, verbose):
    """
    Create transformed image (lower right quadrant for even-only,
    right half for odd) from its cos^n theta radial profiles.
    """
    rbin, wl, wu, cos = _get_image_bs(height, width, row, verbose)

    # 0th order (isotropic)
    IM = (wl * np.append(c[0], [0])[rbin] +  # lower bins
          wu * np.append(c[0][1:], [0, 0])[rbin])  # upper bins
    # add all other orders
    for cn, cosn in zip(c[1:], cos[1:]):
        IM += (wl * np.append(cn, [0])[rbin] +
               wu * np.append(cn[1:], [0, 0])[rbin]) * cosn
    # (weighting for each order is somehow faster than processing lower and
    #  upper bins separately and then combining)

    return IM


def _bs_rbasex(Rmax, order, odd):
    """
    Compute radial parts of basis projections for R and radii up to Rmax.
    """
    # all needed orders (even or all) from 0 to order
    orders = range(0, order + 1, 1 if odd else 2)

    # list of matrces for projections: P[i][R, r] = p_{R:n_i}(r),
    # sampled at r = 0, ..., R, for R = 0, ..., Rmax and orders n
    P = [np.eye(Rmax + 1, order='F') for n in orders]  # ('F' is faster here)
    # (p_{0;0}(0) = 1 indeed, but p_{0;n}(0) = 0 for all other orders,
    #  so P[i][0, 0] are also set to 1 to make P[i] nondegenerate)

    # fill p_{R>0;0}(0) = 2 (all other p_{R>0;n}(0) = 0 already)
    P[0][1:, 0] = 2

    # fill all other r > 0 columns (only functions with R >= r are non-zero)
    for r in range(1, Rmax + 1):
        # all needed R - 1, R, R + 1 for current r
        R = np.arange(r - 1, Rmax + 2, dtype=float)

        # rho = max(r, R)
        rho = R.copy()
        rho[0] = r

        # since z = sqrt(R^2 - r^2) for R >= r, otherwise 0,
        # it is z = sqrt(max(r, R)^2 - r^2) = sqrt(rho^2 - r^2)
        z = np.sqrt(rho**2 - r**2)

        f = r / rho  # "f" means "fraction r / rho"

        rln = r * np.log(z + rho)

        # define F_n for all needed n
        F = {-1: (z / f + rln) / 2, 0: z}
        if order >= 1:
            F[1] = rln
        if order >= 2:
            F[2] = r * np.arccos(f)
        if order >= 3:
            F[3] = z * f
        if order >= 4:
            fn = f.copy()  # current (r / rho)^n
            for n in range(2, order - 1):  # from 4 - 2 to order - 2
                fn *= f
                F[n + 2] = (z * fn + (n - 1) * F[n]) / n

        # compute p_{R;n}(r) for all needed R and n
        for i, n in enumerate(orders):
            rFRF = r * F[n - 1] - R * F[n]
            P[i][r:, r] = 2 * (2 * rFRF[1:-1] - rFRF[2:] - rFRF[:-2])
            #    R >= r                 at R         R - 1      R + 1

    return P


def _load_bs(basis_dir, Rmax, order, odd, inv=False, verbose=False):
    """
    Try to load the basis set and inverse transform matrices from the best
    suitable file.
    Returns (bs, tri), where tri might be None, or (None, None) on failure.
    """
    if basis_dir is None:
        return None, None

    def full_path(file_name):
        return os.path.join(basis_dir, file_name)

    # exact file name
    basis_file = 'rbasex_basis_{}_{}{}{}.npy'.\
                 format(Rmax, order, 'o' if odd else '', 'i' if inv else '')
    if os.path.exists(full_path(basis_file)):
        # have exactly needed
        best_file = basis_file
        best_prm = {'Rmax': Rmax, 'order': order, 'odd': odd, 'inv': inv}
    else:
        # Find the best (smallest among sufficient)
        best_file = None
        best_size = np.inf
        mask = re.compile(r'rbasex_basis_(\d+)_(\d+)(o{})(i?)\.npy'.
                          format('' if odd else '?'))
        for f in listdir(basis_dir):
            # filter rBasex files
            match = mask.match(f)
            if not match:
                continue
            # extract file parameters
            f_Rmax, f_order = map(int, match.groups()[:2])
            f_odd, f_inv = map(bool, match.groups()[2:])
            # skip insufficient files
            if f_Rmax < Rmax or f_order < order:
                continue
            # estimate total size (elements)
            size = f_Rmax**2 * f_order // (1 if f_odd else 2)
            # empirical penalty for no inverse when it is needed
            if inv and not f_inv:
                size *= Rmax
            # skip files larger than sufficient
            if size > best_size:
                continue
            # remember the best so far
            best_file = f
            best_size = size
            best_prm = {'Rmax': f_Rmax, 'order': f_order,
                        'odd': f_odd, 'inv': f_inv}

    if best_file is None:
        return None, None

    if verbose:
        print('Loading basis set from', best_file)
    try:
        bs = np.load(full_path(best_file))
    except ValueError:
        print('Cached basis file incompatible!')
        return None, None

    # pick orders parity
    if best_prm['odd'] > odd:  # odd present but not needed
        bs = bs[::2]  # take only even
        if verbose:
            print('(odd orders skipped)')
    # crop orders
    if best_prm['order'] > order:
        n = 1 + (order if odd else order // 2)
        bs = bs[:n]
        if verbose:
            print('(higher orders skipped)')
    # crop to Rmax
    if best_prm['Rmax'] > Rmax:
        bs = [M[:Rmax + 1, :Rmax + 1] for M in bs]
        if verbose:
            print('(cropped to {})'.format(Rmax))
    # separate into P and Ai
    if best_prm['inv']:
        if inv:
            tri = []
            for n in range(len(bs)):
                # extract upper triangular part
                M = np.triu(bs[n])
                # invert diagonal (it corresponds to P[n])
                M[np.diag_indices_from(M)] = 1 / np.diag(M)
                # store Ai[n] = inv(P[n].T)
                tri.append(M)
                # leave only lower triangular part for P[n]
                bs[n] = np.tril(bs[n])
        else:  # inverse not needed
            for n in range(len(bs)):
                bs[n] = np.tril(bs[n])  # keep only lower triangular part Pn
            tri = None
    else:
        tri = None

    return bs, tri


def _save_bs(basis_dir, Rmax, order, odd, bs, tri=False, verbose=False):
    """
    Try to save the basis set and, if needed, the inverse transform matrix.
    """
    if basis_dir is None:
        return

    has_odd = 'o' if odd else ''
    if tri is not None:
        has_inv = 'i'
        out = []
        for P, Ai in zip(bs, tri):
            # combine lower triangular P with upper triangular Ai without main
            # diagonal (restored on load as diag(Ai) = 1 / diag(P))
            M = np.triu(Ai, 1)
            M += P
            out.append(M)
    else:
        has_inv = ''
        out = bs
    file_name = 'rbasex_basis_{}_{}{}{}.npy'.format(Rmax, order,
                                                    has_odd, has_inv)
    if verbose:
        print('Saving basis set to disk as', file_name)
    np.save(os.path.join(basis_dir, file_name), out)


def get_bs_cached(Rmax, order=2, odd=False, direction='inverse', reg=None,
                  valid=None, basis_dir=None, verbose=False):
    """
    Internal function.

    Gets the basis set (from cache or runs computations and caches them)
    and calculates the transform matrix.
    Loaded/calculated matrices are also cached in memory.

    Parameters
    ----------
    Rmax : int
        largest radius to be transformed
    order : int
        highest angular order
    odd : bool
        include odd angular orders
    direction : str: ``'forward'`` or ``'inverse'``
        type of Abel transform to be performed
    reg : None or str or tuple (str, float)
        regularization type and strength for inverse transform
    valid : None or bool array
        flags to exclude invalid radii from transform
    basis_dir : str, optional
        path to the directory for saving / loading the basis set. Use ``''``
        for the default directory. If ``None``, the basis set will not be
        loaded from or saved to disk.
    verbose : bool
        print some debug information

    Returns
    -------
    A : list of 2D numpy arrays
        (**Rmax** + 1) × (**Rmax** + 1) matrices of the Abel transform (forward
        or inverse) for each angular order
    """
    global _bs_prm, _bs, _trf, _tri_full, _tri_prm, _tri

    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=True)

    new_bs = False  # new basis set computed (for saving to disk)

    prm = [Rmax, order, odd]
    if _bs is None or _bs_prm != prm:
        _bs_prm = prm
        # try to load basis set and maybe inverse-transform matrices
        _bs, _tri_full = _load_bs(basis_dir, Rmax, order, odd,
                                  direction == 'inverse' and reg is None,
                                  verbose)
        if _bs is None:
            if verbose:
                print('Computing basis set...')
            _bs = _bs_rbasex(Rmax, order, odd)
            new_bs = True
        # reset transforms
        _trf = None
        _tri_prm = None
        _tri = None
    else:
        if verbose:
            print('Using cached basis set')

    if valid is None or valid.all():
        invalid = None
    else:
        invalid = np.logical_not(valid)

    def mask(A):
        # Zero rows for output radii without data (columns do not need to be
        # zeroed, since input profiles already have zeros there).
        # Array is modified; to preserve the original — pass a copy.
        if invalid is not None:
            A[invalid] = 0
        return A

    def Af():
        # Make optionally masked forward-transform matrices.
        if invalid is None:
            return [Pn.T for Pn in _bs]
        else:
            return [mask(Pn.T.copy()) for Pn in _bs]

    if direction == 'forward':
        if _trf is None:
            if new_bs:
                _save_bs(basis_dir, Rmax, order, odd, _bs, None, verbose)
            if verbose:
                print('Creating forward-transform matrices...')
            _trf = Af()
        return _trf
    else:  # 'inverse'
        if _tri_prm != [reg]:
            _tri_prm = [reg]
            if reg is None:
                # calculate full inverse matrices, if not yet
                if _tri_full is None:
                    if verbose:
                        print('Calculating inverse-transform matrices...')
                    # P[n] are triangular, thus can be inverted faster than
                    # general matrices, however, NumPy/SciPy do not have such
                    # functions; nevertheless, solve_triangular() is ~twice
                    # faster than inv() (and ...(Pn, I, lower=True).T is faster
                    # than ...(Pn.T, I))
                    I = np.eye(Rmax + 1)
                    _tri_full = [solve_triangular(Pn, I, lower=True).T
                                 for Pn in _bs]
                    new_bs = True
                # mask invalid radii
                _tri = [mask(An.copy()) for An in _tri_full]
            elif reg == 'pos':  # non-negative cos sin
                if verbose:
                    print('Preparing matrices for NNLS equations...')
                # Construct forward transform matrix cossin → cos projections.
                # Notes:
                # 1. By reversing orders, it also could be made triangular for
                #    more effective inversion, but nnls() does not care.
                # 2. This code is not optimized, but its execution time is
                #    still negligible compared to nnls().
                if odd:
                    if order > 1:
                        raise ValueError('reg="pos" is not implemented for '
                                         'odd orders > 1')
                    # use (1 ± cos) / 2
                    A0, A1 = Af()
                    A = [[A0, A0], [A1, -A1]]
                else:  # even only
                    N = 1 + order // 2
                    # cossin → cos transform
                    C = np.flip(invpascal(N, 'upper'))
                    # blocks for each order combination
                    A = [[C[n, m] * An for m in range(N)]
                         for n, An in enumerate(Af())]
                # make single matrix from blocks
                _tri = np.block(A)
            elif np.ndim(reg) == 0:  # not sequence type
                raise ValueError('Wrong regularization format "{}"'.
                                 format(reg))
            elif reg[0] == 'L2':  # Tikhonov L2 norm
                if verbose:
                    print('Calculating L2-regularized transform matrices...')
                E = np.diag([reg[1]] * (Rmax + 1))
                # regularized inverse for each angular order
                _tri = [An.T.dot(inv((An).dot(An.T) + E)) for An in Af()]
            elif reg[0] == 'diff':  # Tikhonov derivative
                if verbose:
                    print('Calculating diff-regularized transform matrices...')
                # GTG = reg D^T D, where D is 1st-order difference operator
                GTG = 2 * np.eye(Rmax + 1) - \
                          np.eye(Rmax + 1, k=-1) - \
                          np.eye(Rmax + 1, k=1)
                GTG[0, 0] = 1
                GTG[-1, -1] = 1
                GTG *= reg[1]
                # regularized inverse for each angular order
                _tri = [An.T.dot(inv((An).dot(An.T) + GTG)) for An in Af()]
            elif reg[0] == 'SVD':
                if verbose:
                    print('Calculating SVD-regularized transform matrices...')
                if reg[1] > 1:
                    raise ValueError('Wrong SVD truncation factor {} > 1'.
                                     format(reg[1]))
                # truncation index (smallest SV of P -> largest SV of inverse)
                smax = int((1 - reg[1]) * Rmax) + 1
                _tri = []
                # loop over angular orders
                for An in Af():
                    U, s, Vh = svd(An.T)
                    # truncate matrices
                    U = U[:, :smax]
                    s = 1 / s[:smax]  # inverse
                    Vh = Vh[:smax]
                    # regularized inverse for this angular order
                    _tri.append((U * s).dot(Vh))
            else:
                raise ValueError('Wrong regularization type "{}"'.
                                 format(reg[0]))
        if new_bs:
            _save_bs(basis_dir, Rmax, order, odd, _bs, _tri_full, verbose)
        return _tri


def cache_cleanup(select='all'):
    """
    Utility function.

    Frees the memory caches created by ``get_bs_cached()``.
    This is usually pointless, but might be required after working
    with very large images, if more RAM is needed for further tasks.

    Parameters
    ----------
    select : str
        selects which caches to clean:

        ``all`` (default)
            everything, including basis;
        ``forward``
            forward transform;
        ``inverse``
            inverse transform.

    Returns
    -------
    None
    """
    global _prm, _dst, _bs_prm, _bs, _ibs, _trf, _tri_full, _tri_prm, _tri

    if select == 'all':
        _prm = None
        _dst = None
        _bs_prm = None
        _bs = None
        _ibs = None
    if select in ('all', 'forward'):
        _trf = None
    if select in ('all', 'inverse'):
        _tri_full = None
        _tri_prm = None
        _tri = None


def basis_dir_cleanup(basis_dir=''):
    """
    Utility function.

    Deletes basis sets saved on disk.

    Parameters
    ----------
    basis_dir : str or None
        absolute or relative path to the directory with saved basis sets. Use
        ``''`` for the default directory. ``None`` does nothing.

    Returns
    -------
    None
    """
    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=False)

    if basis_dir is None:
        return

    files = glob(os.path.join(basis_dir, 'rbasex_basis_*.npy'))
    for fname in files:
        os.remove(fname)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import numpy as np
import abel

dasch_transforms = {
 "two_point": abel.dasch.two_point_transform,
 "three_point": abel.dasch.three_point_transform,
 "onion_peeling": abel.dasch.onion_peeling_transform
}


def test_dasch_shape():
    n = 21
    x = np.ones((n, n), dtype='float32')

    for method in dasch_transforms.keys():
        recon = dasch_transforms[method](x, direction='inverse')
        np.testing.assert_equal(recon.shape, (n, n))


def test_dasch_zeros():
    n = 21
    x = np.zeros((n, n), dtype='float32')

    for method in dasch_transforms.keys():
        recon = dasch_transforms[method](x, direction="inverse")
        np.testing.assert_allclose(recon, 0)


def test_dasch_deconvolution_array_sources():
    im = abel.tools.analytical.SampleImage(101).image
    q = abel.tools.symmetry.get_image_quadrants(im)[0]

    # clean up any old deconvolution array files
    fn = 'three_point_basis_{}.npy'.format(q.shape[0])
    if os.path.exists(fn):
        os.system('rm {}'.format(fn))

    gb = abel.dasch.three_point_transform(q)
    np.testing.assert_equal(abel.dasch._source, 'generated')

    cb = abel.dasch.three_point_transform(q)
    np.testing.assert_equal(abel.dasch._source, 'cache')
    np.testing.assert_allclose(gb, cb)

    abel.dasch.cache_cleanup()
    fb = abel.dasch.three_point_transform(q)
    np.testing.assert_equal(abel.dasch._source, 'file')
    np.testing.assert_allclose(gb, fb)


def test_dasch_1d_gaussian(n=101):
    def gauss(r, r0, sigma):
        return np.exp(-(r-r0)**2/sigma**2)

    rows, cols = n, n
    r2 = rows//2
    c2 = cols//2

    sigma = 20*n/100

    # 1D Gaussian -----------
    r = np.linspace(0, c2-1, c2)

    orig = gauss(r, 0, sigma)

    for method in dasch_transforms.keys():
        orig_copy = orig.copy()

        recon = dasch_transforms[method](orig)

        ratio_1d = np.sqrt(np.pi)*sigma

        np.testing.assert_allclose(orig_copy[20:], recon[20:]*ratio_1d,
                                   rtol=0.0, atol=0.5)


def test_dasch_cyl_gaussian(n=101):
    def gauss(r, r0, sigma):
        return np.exp(-(r-r0)**2/sigma**2)

    image_shape = (n, n)
    rows, cols = image_shape
    r2 = rows//2
    c2 = cols//2
    sigma = 20*n/100

    x = np.linspace(-c2, c2, cols)
    y = np.linspace(-r2, r2, rows)

    X, Y = np.meshgrid(x, y)

    IM = gauss(X, 0, sigma)  # cylindrical Gaussian located at pixel R=0
    Q0 = IM[:r2, c2:]  # quadrant, top-right
    Q0_copy = Q0.copy()
    ospeed = abel.tools.vmi.angular_integration_3D(Q0_copy, origin=(0, 0))

    # dasch method inverse Abel transform
    for method in dasch_transforms.keys():
        Q0_copy = Q0.copy()
        AQ0 = dasch_transforms[method](Q0)
        ratio_2d = np.sqrt(np.pi)*sigma

        np.testing.assert_allclose(Q0_copy, AQ0*ratio_2d, rtol=0.0, atol=0.3)


if __name__ == "__main__":
    test_dasch_shape()
    test_dasch_zeros()
    test_dasch_deconvolution_array_sources()
    test_dasch_1d_gaussian()
    test_dasch_cyl_gaussian()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from .math import fit_gaussian
import warnings
from scipy.ndimage import center_of_mass, shift
from scipy.optimize import minimize
# testing strings with Python 2 and 3 compatibility
from six import string_types

from abel import _deprecated, _deprecate


def find_origin(IM, method='image_center', axes=(0, 1), verbose=False,
                **kwargs):
    """
    Find the coordinates of image origin, using the specified method.

    Parameters
    ----------
    IM : 2D np.array
        image data

    method : str
        determines how the origin should be found. The options are:

        ``image_center``
            the center of the image is used as the origin. The trivial result.
        ``com``
            the origin is found as the center of mass.
        ``convolution``
            the origin is found as the maximum of autoconvolution of the image
            projections along each axis.
        ``gaussian``
            the origin is extracted by a fit to a Gaussian function.
            This is probably only appropriate if the data resembles a
            gaussian.
        ``slice``
            the image is broken into slices, and these slices compared
            for symmetry.

    axes : int or tuple of int
        find origin coordinates: ``0`` (vertical), or ``1`` (horizontal), or
        ``(0, 1)`` (both vertical and horizontal).

    Returns
    -------
    out : (float, float)
        coordinates of the origin of the image in the (row, column) format.
        For coordinates not in **axes**, the center of the image is returned.

    """
    return func_method[method](IM, axes, verbose=verbose, **kwargs)


def center_image(IM, method='com', odd_size=True, square=False, axes=(0, 1),
                 crop='maintain_size', order=3, verbose=False,
                 center=_deprecated, **kwargs):
    """
    Center image with the custom value or by several methods provided in
    :func:`find_origin()` function.

    Parameters
    ----------
    IM : 2D np.array
        The image data.

    method : str or tuple of float
        either a tuple (float, float), the coordinate of the origin of the
        image in the (row, column) format, or a string to specify an automatic
        centering method:

        ``image_center``
            the center of the image is used as the origin. The trivial result.
        ``com``
            the origin is found as the center of mass.
        ``convolution``
            the origin is found as the maximum of autoconvolution of the image
            projections along each axis.
        ``gaussian``
            the origin is extracted from a fit to a Gaussian function.
            This is probably only appropriate if the data resembles a
            gaussian.
        ``slice``
            the image is broken into slices, and these slices compared for
            symmetry.

    odd_size : boolean
        if ``True``, the returned image will contain an odd number of columns.
        Most of the transform methods require this, so it's best to set this
        to ``True`` if the image will subsequently be Abel-transformed.

    square : bool
        if ``True``, the returned image will have a square shape.

    crop : str
        determines how the image should be cropped. The options are:

        ``maintain_size``
            return image of the same size. Some regions of the original image
            may be lost, and some regions may be filled with zeros.
        ``valid_region``
            return the largest image that can be created without padding.
            All of the returned image will correspond to the original image.
            However, portions of the original image will be lost.
            If you can tolerate clipping the edges of the image, this is
            probably the method to choose.
        ``maintain_data``
            the image will be padded with zeros such that none of the original
            image will be cropped.

        See :func:`set_center` for examples.

    axes : int or tuple of int
        center image with respect to axis ``0`` (vertical), ``1`` (horizontal),
        or both axes ``(0, 1)`` (default). When specifying an explicit origin
        in **method**, unused coordinates can also be passed as ``None``, for
        example, ``method=(row, None)`` or ``method=(None, col)``.

    order : int
        interpolation order, see :func:`set_center` for details.

    Returns
    -------
    out : 2D np.array
        centered image
    """
    if center is not _deprecated:
        _deprecate('abel.tools.center.center_image() '
                   'argument "center" is deprecated, use "method" instead.')
        method = center

    rows, cols = IM.shape

    if odd_size and cols % 2 == 0:
        # drop rightside column
        IM = IM[:, :-1]
        rows, cols = IM.shape

    if square and rows != cols:
        # make rows == cols, but maintain approx. center
        if rows > cols:
            diff = rows - cols
            trim = diff // 2
            if trim > 0:
                IM = IM[trim: -trim]  # remove even number of rows off each end
            if diff % 2:
                IM = IM[: -1]  # remove one additional row

        else:
            # make rows == cols, check row oddness
            if odd_size and rows % 2 == 0:
                IM = IM[:-1, :]
                rows -= 1
            xs = (cols - rows) // 2
            IM = IM[:, xs:-xs]

        rows, cols = IM.shape

    # origin is in (row, column) format!
    if isinstance(method, string_types):
        origin = find_origin(IM, method=method, axes=axes, verbose=verbose,
                             **kwargs)
    else:
        origin = method

    centered_data = set_center(IM, origin=origin, crop=crop, axes=axes,
                               order=order, verbose=verbose)
    return centered_data


def set_center(data, origin, crop='maintain_size', axes=(0, 1), order=3,
               verbose=False, center=_deprecated):
    """
    Move image origin to mid-point of image (``rows // 2, cols // 2``).

    Parameters
    ----------
    data : 2D np.array
        the image data

    origin : tuple of float
        (row, column) coordinates of the image origin.
        Coordinates set to ``None`` are ignored.

    crop : str
        determines how the image should be cropped. The options are:

        ``maintain_size`` (default)
            return image of the same size. Some regions of the original image
            may be lost and some regions may be filled with zeros.
        ``valid_region``
            return the largest image that can be created without padding.
            All of the returned image will correspond to the original image.
            However, portions of the original image will be lost.
            If you can tolerate clipping the edges of the image, this is
            probably the method to choose.
        ``maintain_data``
            the image will be padded with zeros such that none of the original
            image will be cropped.

        Examples:

        .. plot:: tools/crop_options.py

    axes : int or tuple of int
        center image with respect to axis ``0`` (vertical), ``1`` (horizontal),
        or both axes ``(0, 1)`` (default).

    order : int
        interpolation order (0–5, default is 3) for centering with fractional
        **origin**. Lower orders work faster; **order** = 0 (also implied for
        integer **origin**) means a whole-pixel shift without interpolation and
        is much faster.

    verbose : bool
        print some information for debugging

    Returns
    -------
    out : 2D np.array
        centered image
    """
    if center is not _deprecated:
        _deprecate('abel.tools.center.set_center() '
                   'argument "center" is deprecated, use "origin" instead.')
        origin = center

    def center_of(a):
        """ Indices of array center """
        return np.array(a.shape) // 2

    shape = data.shape
    if verbose:
        print('Original shape', shape, 'and origin', tuple(origin))
    center = center_of(data)

    # remove axes with "None" coordinates, preprocess origin
    if isinstance(axes, int):
        axes = [axes]
    axes = set(axes)
    origin = np.array(origin, dtype=object)  # (for None, int or float)
    subpixel = np.zeros(2)
    origin_ = [None, None]
    for a in [0, 1]:
        if origin[a] is None:
            axes.discard(a)
        else:
            # to absolute coordinates
            if origin[a] < 0:
                origin[a] += shape[a]
            if order:
                # split integer and fractional parts
                i = int(origin[a])
                origin[a], subpixel[a] = i, origin[a] - i
            else:
                # round to whole pixels
                origin[a] = int(round(origin[a]))
            # complement (from the other edge)
            origin_[a] = shape[a] - 1 - origin[a]
    # don't interpolate for whole-pixels shifts
    if np.all(subpixel == 0):
        order = 0
    if verbose:
        print('Centering axes', tuple(axes), 'using order', order)

    # Note: current scipy.ndimage.shift() implementation erases fractional edge
    # pixels, so we wrap it with padding and cropping. Once this behavior is
    # corrected, our code can be cleaned up.

    if crop == 'maintain_size':
        delta = [0, 0]
        if order:  # fractional shift
            for a in axes:
                if origin[a] is not None:
                    delta[a] = center[a] - (origin[a] + subpixel[a])
            out = shift(np.pad(data, 1, 'constant'),
                        delta, order=order)[1:-1, 1:-1]  # (see note above)
        else:  # whole-pixel shift
            src = [slice(None), slice(None)]  # for the source region
            dst = [slice(None), slice(None)]  # for the destination region
            for a in axes:
                delta[a] = center[a] - origin[a]
                # gaps for positive and negative shifts wrt edges
                dpos = max(0, delta[a])
                dneg = max(0, -delta[a])
                # corresponding regions
                src[a] = slice(dneg, shape[a] - dpos)
                dst[a] = slice(dpos, shape[a] - dneg)
            out = np.zeros_like(data)
            out[tuple(dst)] = data[tuple(src)]
        if verbose:
            print('Shifted by', tuple(delta))
            print('Output shape', out.shape,
                  'centered at', tuple(center_of(out)))
        return out
    # for other crop options, first do subpixel shift
    # size will change to add/remove the shifted fractional pixel parts
    if order:
        if verbose:
            print('Subpixel shift by', tuple(-subpixel))
        # (see the note above about padding on both sides and cropping)
        data = shift(np.pad(data, 1, 'constant'),
                     -subpixel, order=order)[:-1, :-1]
        # shift origin or cut unused pixels
        cut = [slice(None), slice(None)]
        for a in [0, 1]:
            if subpixel[a]:
                if crop == 'valid_region':
                    cut[a] = slice(1, -1)  # cut both fractional ends
                    origin_[a] -= 1
                else:
                    origin[a] += 1
                # (shape is not used below, thus not updated here)
            else:
                cut[a] = slice(1, None)  # cut empty (not shifted) pixel
        data = data[tuple(cut)]
    if crop == 'valid_region':
        src = [slice(None), slice(None)]
        for a in axes:
            # distance to the closest edge
            d = min(origin[a], origin_[a])
            # crop symmetrically around the origin
            src[a] = slice(origin[a] - d, origin[a] + d + 1)
        out = data[tuple(src)].copy()  # (independent data, as in other cases)
        if verbose:
            print('Output cropped to', out.shape,
                  'centered at', tuple(center_of(out)))
        return out
    elif crop == 'maintain_data':
        pad = [(0, 0), (0, 0)]
        for a in axes:
            # distance to the farthest edge
            d = max(origin[a], origin_[a])
            # pad to symmetrize around the origin
            pad[a] = (d - origin[a], d - origin_[a])
        out = np.pad(data, pad, 'constant')
        if verbose:
            print('Output padded to', out.shape,
                  'centered at', tuple(center_of(out)))
        return out
    else:
        raise ValueError('Invalid crop option "{}".'.format(crop))


def find_origin_by_center_of_mass(IM, axes=(0, 1), verbose=False,
                                  round_output=False, **kwargs):
    """
    Find image origin by calculating its center of mass.

    Parameters
    ----------
    IM : numpy 2D array
        image data

    round_output : bool
        if ``True``, the coordinates are rounded to integers;
        otherwise they are floats.

    axes : int or tuple
        find origin coordinates: ``0`` (vertical), or ``1`` (horizontal), or
        ``(0, 1)`` (both vertical and horizontal).

    Returns
    -------
    origin : (float, float)
        (row, column)
    """
    origin = list(center_of_mass(IM))

    # reset unneeded coordinates
    if isinstance(axes, int):
        axes = [axes]
    for a in set([0, 1]) - set(axes):
        origin[a] = IM.shape[a] // 2
    origin = tuple(origin)

    if verbose:
        to_print = "Center of mass at {}".format(origin)

    if round_output:
        origin = (round(origin[0]), round(origin[1]))
        if verbose:
            to_print += " ... round to {}".format(origin)

    if verbose:
        print(to_print)

    return origin


def find_origin_by_convolution(IM, axes=(0, 1), projections=False, **kwargs):
    """
    Find the image origin as the maximum of autoconvolution of its projections
    along each axis.

    Code from the ``linbasex`` juptyer notebook.

    Parameters
    ----------
    IM : numpy 2D array
        image data

    projections : bool
        if this parameter is ``True``, the autoconvoluted projections along
        both axes will be returned after the origin.

    axes : int or tuple
        find origin coordinates: ``0`` (vertical), or ``1`` (horizontal), or
        ``(0, 1)`` (both vertical and horizontal).

    Returns
    -------
    origin : (float, float)
        (row, column)

        `or` (row, column), conv_0, conv_1
    """
    if isinstance(axes, int):
        axes = [axes]

    conv = [None, None]
    origin = [IM.shape[0] // 2, IM.shape[1] // 2]
    for a in axes:
        # projection along the other axis
        proj = IM.sum(axis=1 - a)
        # autoconvolute projections
        conv[a] = np.convolve(proj, proj, mode='full')
        # take the first max, should there be several equal maxima
        origin[a] = np.argmax(conv[a]) / 2
    origin = tuple(origin)

    if projections:
        return origin, conv[0], conv[1]
    else:
        return origin


def find_origin_by_center_of_image(data, axes=(0, 1), verbose=False, **kwargs):
    """
    Find image origin simply as its center, from its dimensions.

    Parameters
    ----------
    IM : numpy 2D array
        image data

    axes : int or tuple
        has no effect

    Returns
    -------
    origin : (int, int)
        (row, column)

    """
    return (data.shape[0] // 2, data.shape[1] // 2)


def find_origin_by_gaussian_fit(IM, axes=(0, 1), verbose=False,
                                round_output=False, **kwargs):
    """
    Find image origin by fitting the summation along rows and columns of the
    data to two 1D Gaussian functions.

    Parameters
    ----------
    IM : numpy 2D array
        image data

    axes : int or tuple
        find origin coordinates: ``0`` (vertical), or ``1`` (horizontal), or
        ``(0, 1)`` (both vertical and horizontal).

    round_output : bool
        if ``True``, the coordinates are rounded to integers;
        otherwise they are floats.

    Returns
    -------
    origin : (float, float)
        (row, column)
    """
    if isinstance(axes, int):
        axes = [axes]

    origin = [IM.shape[0] // 2, IM.shape[1] // 2]
    for a in axes:
        # sum along the other axis
        proj = np.sum(IM, axis=1 - a)
        # find gaussian center
        origin[a] = fit_gaussian(proj)[1]
    origin = tuple(origin)

    if verbose:
        to_print = "Gaussian origin at {}".format(origin)

    if round_output:
        origin = (round(origin[0]), round(origin[1]))
        if verbose:
            to_print += " ... round to {}".format(origin)

    if verbose:
        print(to_print)

    return origin


def axis_slices(IM, radial_range=(0, -1), slice_width=10):
    """
    Returns vertical and horizontal slice profiles, summed across slice_width.

    Parameters
    ----------
    IM : 2D np.array
        image data

    radial_range: tuple of float
        (rmin, rmax) range to limit data

    slice_width : int
        width of the image slice, default 10 pixels

    Returns
    -------
    top, bottom, left, right : 1D np.arrays shape (rmin:rmax, 1)
        image slices oriented in the same direction
    """
    rows, cols = IM.shape   # image size

    r2 = rows // 2
    c2 = cols // 2
    sw2 = min(slice_width // 2, r2, c2)

    rmin, rmax = radial_range

    # vertical slice
    top = IM[:r2, c2-sw2:c2+sw2+1].sum(axis=1)
    bottom = IM[r2 + rows % 2:, c2-sw2:c2+sw2+1].sum(axis=1)

    # horizontal slice
    left = IM[r2-sw2:r2+sw2+1, :c2].sum(axis=0)
    right = IM[r2-sw2:r2+sw2+1, c2 + cols % 2:].sum(axis=0)

    return (top[::-1][rmin:rmax], bottom[rmin:rmax],
            left[::-1][rmin:rmax], right[rmin:rmax])


def find_origin_by_slice(IM, axes=(0, 1), slice_width=10, radial_range=(0, -1),
                         axis=_deprecated, **kwargs):
    """
    Find the image origin by comparing opposite sides.

    Parameters
    ----------
    IM : 2D np.array
        the image data

    slice_width : integer
        Sum together this number of rows (cols) to improve signal, default 10.

    radial_range: tuple
        (rmin, rmax): radial range ``[rmin:rmax]`` for slice profile
        comparison.

    axes : int or tuple
        find origin coordinates: ``0`` (vertical), or ``1`` (horizontal), or
        ``(0, 1)`` (both vertical and horizontal).

    Returns
    -------
    origin : (float, float)
        (row, column)

    """
    if axis is not _deprecated:
        _deprecate('abel.tools.center.find_origin_by_slice() '
                   'argument "axis" is deprecated, use "axes" instead.')
        axes = axis

    def _align(offset, sliceA, sliceB):
        """
        Intensity difference between an axial slice and its shifted opposite.
        """
        # always shift to the left (towards center)
        if offset < 0:
            diff = shift(sliceA, offset) - sliceB
        else:
            diff = sliceA - shift(sliceB, -offset)
        fvec = (diff**2).sum()
        return fvec

    if isinstance(axes, int):
        axes = [axes]

    rows, cols = IM.shape

    r2 = (rows - 1) / 2
    c2 = (cols - 1) / 2
    top, bottom, left, right = axis_slices(IM, radial_range, slice_width)

    xyoffset = [0.0, 0.0]
    # determine shift to align both slices
    # limit shift to +- 20 pixels
    initial_shift = [0.1, ]

    # vertical axis
    if 0 in axes:
        fit = minimize(_align, initial_shift, args=(top, bottom),
                       bounds=((-50, 50), ), tol=0.1)
        if fit["success"]:
            xyoffset[0] = -float(fit['x']) / 2  # x1/2 for image shift
        else:
            raise RuntimeError("fit failure: axis 0, zero shift set", fit)

    # horizontal axis
    if 1 in axes:
        fit = minimize(_align, initial_shift, args=(left, right),
                       bounds=((-50, 50), ), tol=0.1)
        if fit["success"]:
            xyoffset[1] = -float(fit['x']) / 2  # x1/2 for image shift
        else:
            raise RuntimeError("fit failure: axis 1, zero shift set", fit)

    # this is the (row, col) shift to align the slice profiles
    return r2 - xyoffset[0], c2 - xyoffset[1]


func_method = {
    "image_center": find_origin_by_center_of_image,
    "com": find_origin_by_center_of_mass,
    "convolution": find_origin_by_convolution,
    "gaussian": find_origin_by_gaussian_fit,
    "slice": find_origin_by_slice
}


# Deprecated functions

def find_center(IM, center='image_center', square=False, verbose=False,
                **kwargs):
    """Deprecated function. Use :func:`find_origin` instead."""
    _deprecate('abel.tools.center.find_center() '
               'is deprecated, use abel.tools.center.find_origin() instead.')
    if square:
        _deprecate('Argument "square" has no effect and is deprecated.')
    return find_origin(IM, center, verbose, **kwargs)


def find_center_by_center_of_mass(IM, verbose=False, round_output=False,
                                  **kwargs):
    """Deprecated function. Use :func:`find_origin_by_center_of_mass`
    instead."""
    _deprecate('abel.tools.center.find_center_by_center_of_mass() '
               'is deprecated, use '
               'abel.tools.center.find_origin_by_center_of_mass() instead.')
    return find_origin_by_center_of_mass(IM, verbose, round_output, **kwargs)


def find_center_by_convolution(IM, **kwargs):
    """Deprecated function. Use :func:`find_origin_by_convolution` instead."""
    _deprecate('abel.tools.center.find_center_by_convolution() '
               'is deprecated, use '
               'abel.tools.center.find_origin_by_convolution() instead.')
    return find_origin_by_convolution(IM, **kwargs)


def find_center_by_center_of_image(data, verbose=False, **kwargs):
    """Deprecated function. Use :func:`find_origin_by_center_of_image`
    instead."""
    _deprecate('abel.tools.center.find_center_by_center_of_image() '
               'is deprecated, use '
               'abel.tools.center.find_origin_by_center_of_image() instead.')
    return find_origin_by_center_of_image(data, verbose, **kwargs)


def find_center_by_gaussian_fit(IM, verbose=False, round_output=False,
                                **kwargs):
    """Deprecated function. Use :func:`find_origin_by_gaussian_fit` instead."""
    _deprecate('abel.tools.center.find_center_by_gaussian_fit() '
               'is deprecated, use '
               'abel.tools.center.find_origin_by_gaussian_fit() instead.')
    return find_origin_by_gaussian_fit(IM, verbose, round_output, **kwargs)


def find_image_center_by_slice(IM, slice_width=10, radial_range=(0, -1),
                               axis=(0, 1), **kwargs):
    """Deprecated function. Use :func:`find_origin_by_slice` instead."""
    _deprecate('abel.tools.center.find_image_center_by_slice() '
               'is deprecated, use '
               'abel.tools.center.find_origin_by_slice() instead.')
    return find_origin_by_slice(IM, slice_width, radial_range, axis, **kwargs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from numpy.testing import assert_allclose

import abel
from abel.tools.circularize import circularize, circularize_image


def test_circularize_image():

    IM = abel.tools.analytical.SampleImage(n=511, name='Ominus', sigma=2).image

    # flower image distortion
    def flower_scaling(theta, freq=2, amp=0.1):
        return 1 + amp * np.sin(freq * theta)**4

    IMdist = circularize(IM, radial_correction_function=flower_scaling)

    IMcirc, angle, scalefactor, spline = \
        circularize_image(IMdist,
                          method='lsq', dr=0.5, dt=0.1, tol=0,
                          ref_angle=0, return_correction=True)

    r, c = IMcirc.shape

    diff = (IMcirc - IM).sum()

    assert_allclose(diff, -307.603, atol=0.05)

    assert_allclose(angle[-1], 3.04, atol=0.05)

    assert_allclose(scalefactor[4], 0.97, atol=0.05)


if __name__ == "__main__":
    test_circularize_image()



from __future__ import division

import numpy as np
from numpy.testing import assert_equal, assert_allclose
import itertools

from abel.tools.vmi import Distributions, harmonics


def test_origin():
    """
    Test symbolic origin wrt corresponding numeric form.
    """
    # image size
    n = 11  # width
    m = 20  # height
    # origin (symbolic)
    row = 'utclb'
    col = 'lcr'
    # origin (numeric)
    xc = [0, n // 2, n - 1]
    yc = [0, 0, m // 2, m - 1, m - 1]

    np.random.seed(0)
    IM = np.random.random((m, n))

    for r, y in zip(row, yc):
        for c, x in zip(col, xc):
            assert_equal(harmonics(IM, r + c), harmonics(IM, (y, x)),
                         err_msg='-> origin "{}" != {}'.format(r + c, (y, x)))


def run_order(method, tol):
    """
    Test cossin distributions for even orders using Gaussian peaks.

    tol = list of tolerances for order=0, order=2, order=4, ...
    """
    sigma = 5.0  # peak SD
    step = 6 * sigma  # distance between peak centers

    for n in range(len(tol)):  # order = 2n
        size = int((n + 2) * step)
        # squared coordinates:
        x2 = np.arange(float(size))**2
        y2 = x2[:, None]
        r2 = x2 + y2
        # cos^2, sin^2
        c2 = np.divide(y2, r2, out=np.zeros_like(r2), where=r2 != 0)
        s2 = 1 - c2
        # radius
        r = np.sqrt(r2)

        # Gaussian peak with one cossin angular term.
        def peak(i):
            return c2 ** i * s2 ** (n - i) * \
                   np.exp(-(r - (i + 1) * step) ** 2 / (2 * sigma**2))

        # quadrant with all peaks
        Q = peak(0)
        for i in range(1, n + 1):
            Q += peak(i)

        for rmax in ['MIN', 'all']:
            param = '-> rmax = {}, order={}, method = {}'.\
                    format(rmax, 2 * n, method)
            res = Distributions('ul', rmax, 2 * n, method=method).image(Q)
            cossin = res.cossin()
            # extract values at peak centers
            cs = np.array([cossin[:, int((i + 1) * step)]
                           for i in range(n + 1)])
            assert_allclose(cs, np.identity(n + 1), atol=tol[n],
                            err_msg=param)
            # test that Ibeta, and thus harmonics, work in principle
            res.Ibeta()


def test_order_nearest():
    run_order('nearest', [0.0018, 0.0018, 0.0022, 0.0062, 0.020])


def test_order_linear():
    run_order('linear', [0.0031, 0.0034, 0.0034, 0.0067, 0.029])


def test_order_remap():
    run_order('remap', [5e-6, 6e-6, 6e-6, 2e-5, 7e-5])


def run_order_odd(method, tol):
    """
    Test cossin distributions including odd orders using Gaussian peaks.

    tol = list of tolerances for order=0, order=1, order=2, ...
    """
    sigma = 5.0  # peak SD
    step = 6 * sigma  # distance between peak centers

    for n in range(0, len(tol)):  # order = n
        size = int((n + 2) * step)
        # coordinates:
        x = np.arange(float(size))
        y = size - np.arange(2 * float(size) + 1)[:, None]
        # radius
        r = np.sqrt(x**2 + y**2)
        # cos, sin
        r[size, 0] = np.inf
        c = y / r
        s = x / r
        r[size, 0] = 0

        # Gaussian peak with one cossin angular term
        def peak(i):
            m = i  # cos power
            k = (n - m) & ~1  # sin power (round down to even)
            return c ** m * s ** k * \
                   np.exp(-(r - (i + 1) * step) ** 2 / (2 * sigma**2))

        # quadrant with all peaks
        Q = peak(0)
        for i in range(1, n + 1):
            Q += peak(i)

        for rmax in ['MIN', 'all']:
            param = '-> rmax = {}, order={}, method = {}'.\
                    format(rmax, n, method)
            res = Distributions((int(size), 0), rmax, n, odd=True,
                                method=method).image(Q)
            cossin = res.cossin()
            # extract values at peak centers
            cs = np.array([cossin[:, int((i + 1) * step)]
                           for i in range(n + 1)])
            assert_allclose(cs, np.identity(n + 1), atol=tol[n],
                            err_msg=param)
            # test that Ibeta, and thus harmonics, work in principle
            res.Ibeta()


def test_order_odd_nearest():
    run_order_odd('nearest', [0.0018, 0.0018, 0.0018, 0.0018, 0.0020])


def test_order_odd_linear():
    run_order_odd('linear', [0.0032, 0.0034, 0.0034, 0.0034, 0.0035])


def test_order_odd_remap():
    run_order_odd('remap', [5e-6, 5e-6, 6e-6, 6e-6, 6e-6])


def run_method(method, rmax, tolP0, tolP2, tolI, tolbeta, weq=True):
    """
    Test harmonics and Ibeta for various combinations of origins and weights
    for default order=2, odd=False.

    method = method name
    rmax = 'MIN' or 'all'
    tol... = (atol, rmstol) for ...
    tolbeta = atol for beta
    weq = compare symbolic and array weights
    """
    # image size
    n = 81  # width
    m = 71  # height
    # origin coordinates
    xc = [0, 20, n // 2, 60, n - 1]
    yc = [0, 25, m // 2, 45, m - 1]
    # peak SD
    sigma = 2.0

    # Gaussian peak.
    def peak(i, r):
        return np.exp(-(r - i * step)**2 / (2 * sigma**2))

    # Test image.
    def image():
        # squared coordinates:
        x2 = (np.arange(float(n)) - x0)**2
        y2 = (np.arange(float(m))[:, None] - y0)**2
        r2 = x2 + y2
        # radius:
        r = np.sqrt(r2)
        # cos^2, sin^2:
        c2 = np.divide(y2, r2, out=np.zeros_like(r2), where=r2 != 0)
        s2 = 1 - c2
        # image: 3 peaks with different anisotropies
        IM = s2 * peak(1, r) + \
                  peak(2, r) + \
             c2 * peak(3, r)
        return IM, np.sqrt(s2)  # image, sin theta

    # Reference distribution for test image.
    def ref_distr():
        r = np.arange(R + 1)
        P0 = 2/3 * peak(1, r) + \
                   peak(2, r) + \
             1/3 * peak(3, r)
        P2 = -2/3 * peak(1, r) + \
              2/3 * peak(3, r)
        I = 4 * np.pi * r**2 * P0
        beta = P2 / P0
        return r, P0, P2, I, beta

    for y0, x0 in itertools.product(yc, xc):
        param = ' @ y0 = {}, x0 = {}, rmax = {}, method = {}'.\
                format(y0, x0, rmax, method)

        # determine largest radius extracted from image
        if rmax == 'MIN':
            R = min(max(x0, n - 1 - x0), max(y0, m - 1 - y0))
        elif rmax == 'all':
            R = max([int(np.sqrt((x - x0)**2 + (y - y0)**2))
                     for x in (0, n - 1) for y in (0, m - 1)])
        step = (R - 5 * sigma) / 3  # distance between peak centers
        refr, refP0, refP2, refI, refbeta = ref_distr()
        f = 1 / (4 * np.pi * (1 + refr**2))  # rescaling factor for I

        IM, ws = image()
        w1 = np.ones_like(IM)

        IMcopy = IM.copy()
        w1copy = w1.copy()
        wscopy = ws.copy()

        weights = [(False, None, None),
                   (True, None, None),
                   (False, '1', w1),
                   (False, 'sin', ws),
                   (True, '1', w1)]
        P0, P2, r, I, beta = {}, {}, {}, {}, {}
        for use_sin, wname, warray in weights:
            weight_param = param + \
                           ', sin = {}, weights = {}'.format(use_sin, wname)
            key = (use_sin, wname)

            distr = Distributions((y0, x0), rmax, use_sin=use_sin,
                                  weights=warray, method=method)
            res = distr(IM)
            P0[key], P2[key] = res.harmonics()
            r[key], I[key], beta[key] = res.rIbeta()

            def assert_cmp(msg, a, ref, tol):
                atol, rmstol = tol
                assert_allclose(a, ref, atol=atol,
                                err_msg=msg + weight_param)
                rms = np.sqrt(np.mean((a - ref)**2))
                assert rms < rmstol, \
                       '\n' + msg + weight_param + \
                       '\nRMS error = {} > {}'.format(rms, rmstol)

            assert_cmp('-> P0', P0[key], refP0, tolP0)
            assert_cmp('-> P2', P2[key], refP2, tolP2)

            assert_equal(r[key], refr, err_msg='-> r' + weight_param)
            assert_cmp('-> I', f * I[key], f * refI, tolI)
            # beta values at peak centers
            b = [round(beta[key][int(i * step)], 5) for i in (1, 2, 3)]
            assert_allclose(b, [-1, 0, 2], atol=tolbeta,
                            err_msg='-> beta' + weight_param)

            assert_equal(IM, IMcopy,
                         err_msg='-> IM corrupted' + weight_param)
            assert_equal(w1, w1copy,
                         err_msg='-> weights corrupted' + weight_param)
            assert_equal(ws, wscopy,
                         err_msg='-> weights corrupted' + weight_param)

        if not weq:
            continue
        # check that results for symbolic and explicit weights match
        for key1, key2 in [((False, '1'), (False, None)),
                           ((False, 'sin'), (True, None)),
                           ((True, '1'), (False, 'sin'))]:
            pair_param = param + ', sin + weights {} != {}'.format(key1, key2)
            assert_allclose(P0[key1], P0[key2],
                            err_msg='-> P0' + pair_param)
            assert_allclose(P2[key1], P2[key2],
                            err_msg='-> P2' + pair_param)
            assert_allclose(I[key1], I[key2],
                            err_msg='-> I' + pair_param)
            assert_allclose(beta[key1], beta[key2],
                            err_msg='-> beta' + pair_param)


def test_nearest():
    run_method('nearest', 'MIN',
               (0.030, 0.011), (0.035, 0.012), (0.030, 0.011), 0.017)
    run_method('nearest', 'all',
               (0.047, 0.012), (0.071, 0.017), (0.047, 0.012), 0.093)


def test_linear():
    run_method('linear', 'MIN',
               (0.020, 0.0074), (0.016, 0.0049), (0.020, 0.0073), 0.0076)
    run_method('linear', 'all',
               (0.055, 0.012), (0.083, 0.017), (0.055, 0.012), 0.068)


def test_remap():
    run_method('remap', 'MIN',
               (0.0035, 0.00078), (0.0086, 0.0019), (0.0035, 0.00078), 0.0021,
               weq=False)
    run_method('remap', 'all',
               (0.027, 0.0046), (0.041, 0.0071), (0.027, 0.0046), 0.073,
               weq=False)
    # (resampling of weights array in 'remap' makes "weq" differ)


def run_method_odd(method, rmax, tolP0, tolP1, tolP2,
                   tolI, tolbeta1, tolbeta2, weq=True):
    """
    Test harmonics and Ibeta for various combinations of origins and weights
    for default order=2, but with odd=True.

    method = method name
    rmax = 'MIN' or 'all'
    tol... = (atol, rmstol) for ...
    tolbeta = atol for beta
    weq = compare symbolic and array weights
    """
    # image size
    n = 81  # width
    m = 91  # height
    # origin coordinates
    xc = [0, 30, n // 2, 50, n - 1]
    yc = [0, 25, m // 2, 65, m - 1]
    # peak SD
    sigma = 2.0

    # Gaussian peak.
    def peak(i, r):
        return np.exp(-(r - i * step)**2 / (2 * sigma**2))

    # Test image.
    def image():
        # coordinates:
        x = np.arange(float(n)) - x0
        y = y0 - np.arange(float(m))[:, None]
        # radius:
        r = np.sqrt(x**2 + y**2)
        # cos, |sin|
        r[y0, x0] = np.inf
        c = y / r
        s = np.abs(x) / r
        s[y0, x0] = 1
        r[y0, x0] = 0
        # image: 4 peaks with different anisotropies
        IM = s**2      * peak(1, r) + \
             c**2      * peak(2, r) + \
             (1/2 + c) * peak(3, r) + \
                         peak(4, r)
        return IM, s  # image, sin theta

    # Reference distribution for test image.
    def ref_distr():
        r = np.arange(R + 1)
        P0 = 2/3 * peak(1, r) + \
             1/3 * peak(2, r) + \
             1/2 * peak(3, r) + \
                   peak(4, r)
        P1 = peak(3, r)
        P2 = -2/3 * peak(1, r) + \
              2/3 * peak(2, r)
        I = 4 * np.pi * r**2 * P0
        beta1 = P1 / P0
        beta2 = P2 / P0
        return r, P0, P1, P2, I, beta1, beta2

    for y0, x0 in itertools.product(yc, xc):
        param = ' @ y0 = {}, x0 = {}, rmax = {}, method = {}'.\
                format(y0, x0, rmax, method)

        # determine largest radius extracted from image
        if rmax == 'MIN':
            R = min(max(x0, n - 1 - x0), max(y0, m - 1 - y0))
        elif rmax == 'MAX':
            # exclude situations when the outer ring has insufficient vertical
            # span for reliable even/odd separation
            if y0 in [0, m - 1] and x0 not in [0, n - 1] or \
               x0 == n // 2 and abs(y0 - m // 2) not in [0, m // 2]:
                continue
            R = max(max(x0, n - 1 - x0), max(y0, m - 1 - y0))
        step = (R - 5 * sigma) / 4  # distance between peak centers
        refr, refP0, refP1, refP2, refI, refbeta1, refbeta2 = ref_distr()
        f = 1 / (4 * np.pi * (1 + refr**2))  # rescaling factor for I

        IM, ws = image()
        w1 = np.ones_like(IM)

        IMcopy = IM.copy()
        w1copy = w1.copy()
        wscopy = ws.copy()

        weights = [(False, None, None),
                   (True, None, None),
                   (False, '1', w1),
                   (False, 'sin', ws),
                   (True, '1', w1)]
        P0, P1, P2, r, I, beta1, beta2 = {}, {}, {}, {}, {}, {}, {}
        for use_sin, wname, warray in weights:
            weight_param = param + \
                           ', sin = {}, weights = {}'.format(use_sin, wname)
            key = (use_sin, wname)

            distr = Distributions((y0, x0), rmax, odd=True, use_sin=use_sin,
                                  weights=warray, method=method)
            res = distr(IM)
            P0[key], P1[key], P2[key] = res.harmonics()
            r[key], I[key], beta1[key], beta2[key] = res.rIbeta()

            def assert_cmp(msg, a, ref, tol):
                atol, rmstol = tol
                assert_allclose(a, ref, atol=atol,
                                err_msg=msg + weight_param)
                rms = np.sqrt(np.mean((a - ref)**2))
                assert rms < rmstol, \
                       '\n' + msg + weight_param + \
                       '\nRMS error = {} > {}'.format(rms, rmstol)

            assert_cmp('-> P0', P0[key], refP0, tolP0)
            assert_cmp('-> P1', P1[key], refP1, tolP1)
            assert_cmp('-> P2', P2[key], refP2, tolP2)

            assert_equal(r[key], refr, err_msg='-> r' + weight_param)
            assert_cmp('-> I', f * I[key], f * refI, tolI)
            # beta values at peak centers
            b1 = [round(beta1[key][int(i * step)], 5) for i in (1, 2, 3, 4)]
            assert_allclose(b1, [0, 0, 2, 0], atol=tolbeta1,
                            err_msg='-> beta1' + weight_param)
            b2 = [round(beta2[key][int(i * step)], 5) for i in (1, 2, 3, 4)]
            assert_allclose(b2, [-1, 2, 0, 0], atol=tolbeta2,
                            err_msg='-> beta2' + weight_param)

            assert_equal(IM, IMcopy,
                         err_msg='-> IM corrupted' + weight_param)
            assert_equal(w1, w1copy,
                         err_msg='-> weights corrupted' + weight_param)
            assert_equal(ws, wscopy,
                         err_msg='-> weights corrupted' + weight_param)

        if not weq:
            continue
        # check that results for symbolic and explicit weights match
        for key1, key2 in [((False, '1'), (False, None)),
                           ((False, 'sin'), (True, None)),
                           ((True, '1'), (False, 'sin'))]:
            pair_param = param + ', sin + weights {} != {}'.format(key1, key2)
            assert_allclose(P0[key1], P0[key2],
                            err_msg='-> P0' + pair_param)
            assert_allclose(P1[key1], P1[key2],
                            err_msg='-> P1' + pair_param)
            assert_allclose(P2[key1], P2[key2],
                            err_msg='-> P2' + pair_param)
            assert_allclose(I[key1], I[key2],
                            err_msg='-> I' + pair_param)
            assert_allclose(beta1[key1], beta1[key2],
                            err_msg='-> beta1' + pair_param)
            assert_allclose(beta2[key1], beta2[key2],
                            err_msg='-> beta2' + pair_param)


def test_nearest_odd():
    run_method_odd('nearest', 'MIN',
                   (0.093, 0.030), (0.21, 0.063), (0.17, 0.042),
                   (0.093, 0.030), 0.18, 0.075)
    run_method_odd('nearest', 'MAX',
                   (0.16, 0.029), (0.27, 0.049), (0.12, 0.022),
                   (0.16, 0.029), 0.080, 0.030)


def test_linear_odd():
    run_method_odd('linear', 'MIN',
                   (0.025, 0.0078), (0.072, 0.028), (0.054, 0.019),
                   (0.025, 0.0078), 0.17, 0.11)
    run_method_odd('linear', 'MAX',
                   (0.20, 0.040), (0.33, 0.064), (0.14, 0.027),
                   (0.20, 0.040), 0.11, 0.086)


def test_remap_odd():
    run_method_odd('remap', 'MIN',
                   (0.016, 0.0033), (0.029, 0.0060), (0.015, 0.0030),
                   (0.016, 0.0033), 0.050, 0.012, weq=False)
    run_method_odd('remap', 'MAX',
                   (0.13, 0.025), (0.21, 0.041), (0.087, 0.017),
                   (0.13, 0.025), 0.021, 0.011, weq=False)
    # (resampling of weights array in 'remap' makes "weq" differ)


def run_random(method, odd, use_sin, parts=True):
    """
    Test quadrant flipping and image folding using random image data.

    method = method name
    parts = test that masking by weights array equals image cropping
    """
    # image size
    n = 51  # width
    m = 41  # height

    np.random.seed(0)
    IM = np.random.random((m, n)) + 1
    weights = np.random.random((m, n)) + 1

    # quadrant flips
    for y in [0, m - 1]:
        ydir = -1 if y > 0 else 1
        for x in [0, n - 1]:
            xdir = -1 if x > 0 else 1
            ho = harmonics(IM, (y, x), 'all', odd=odd,
                           use_sin=use_sin, weights=weights,
                           method=method)
            hf = harmonics(IM[::ydir, ::xdir], (0, 0), 'all', odd=odd,
                           use_sin=use_sin, weights=weights[::ydir, ::xdir],
                           method=method)
            if odd:
                ho = ho[:, :-3]
                hf = hf[:, :-3]
                if ydir == -1:
                    hf[1] = -hf[1]
                cmp = assert_allclose
            else:
                cmp = assert_equal
            cmp(ho, hf,
                err_msg='-> flip({}, {}) @ method = {}, odd = {}, sin = {}'.
                        format(ydir, xdir, method, odd, use_sin))

    # parts
    if not parts:
        return
    for y0, x0 in itertools.product([15, m // 2, 25], [10, n // 2, 40]):
        param = ' @ y0 = {}, x0 = {}, method = {}, odd = {}, sin = {}'.\
                format(y0, x0, method, odd, use_sin)

        # trim border
        trim = (slice(1, -1), slice(1, -1))
        IMtrim = IM[trim]
        wtrim = weights[trim]
        ht = harmonics(IMtrim, (y0 - 1, x0 - 1), 'all', odd=odd,
                       use_sin=use_sin, weights=wtrim, method=method)

        wmask = np.zeros_like(IM)
        wmask[trim] = weights[trim]
        hm = harmonics(IM, (y0, x0), 'all', odd=odd, use_sin=use_sin,
                       weights=wmask, method=method)

        assert_allclose(ht, hm[:, :ht.shape[1]], err_msg='-> trim' + param)

        # cut quadrants
        regions = [
            ((slice(0, y0 + 1), slice(0, x0 + 1)), 'lr'),
            ((slice(0, y0 + 1), slice(x0, None)),  'll'),
            ((slice(y0, None),  slice(0, x0 + 1)), 'ur'),
            ((slice(y0, None),  slice(x0, None)),  'ul'),
        ]
        for region, origin in regions:
            Q = IM[region]
            Qw = weights[region]
            hc = harmonics(Q, origin, 'all', odd=odd, use_sin=use_sin,
                           weights=Qw, method=method)

            wmask = np.zeros_like(IM)
            wmask[region] = weights[region]
            hm = harmonics(IM, (y0, x0), 'all', odd=odd, use_sin=use_sin,
                           weights=wmask, method=method)

            assert_allclose(hc, hm[:, :hc.shape[1]],
                            err_msg='-> Q (origin = ' + origin + ')' + param)


def test_nearest_random():
    for odd in [False, True]:
        for use_sin in [False, True]:
            run_random('nearest', odd, use_sin)


def test_linear_random():
    for odd in [False, True]:
        for use_sin in [False, True]:
            run_random('linear', odd, use_sin)


def test_remap_random():
    for use_sin in [False, True]:
        run_random('remap', False, use_sin, parts=False)
    # (interpolation and different sampling in 'remap' make
    #  odd vertical flip and "parts" differ)


if __name__ == '__main__':
    test_origin()

    test_order_nearest()
    test_order_odd_nearest()
    test_nearest()
    test_nearest_odd()
    test_nearest_random()

    test_order_linear()
    test_order_odd_linear()
    test_linear()
    test_linear_odd()
    test_linear_random()

    test_order_remap()
    test_order_odd_remap()
    test_remap()
    test_remap_odd()
    test_remap_random()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

#############################################################################
# hansenlaw - a recursive method forward/inverse Abel transform algorithm
#
# Adapted from (see also PR #211):
#  [1] E. W. Hansen "Fast Hankel Transform"
#      IEEE Trans. Acoust. Speech, Signal Proc. 33(3), 666-671 (1985)
#      doi: 10.1109/TASSP.1985.1164579
#
# and:
#
#  [2] E. W. Hansen and P-L. Law
#      "Recursive methods for computing the Abel transform and its inverse"
#      J. Opt. Soc. Am A2, 510-520 (1985)
#      doi: 10.1364/JOSAA.2.000510
#
# 2019-04   : replace gradient with first-order finite difference as
#             per @MikhailRyazanov suggestion
# 2018-04   : New code rewrite, implementing the 1st-order hold approx. of
#             Ref. [1], with the assistance of Eric Hansen. See PR #211.
#
#             Original hansenlaw code was based on Ref. [2]
#
# 2018-03   : NB method applies to grid centered (even columns), not
#             pixel-centered (odd column) image see #206, #211
#             Apply, -1/2 pixel shift for odd column full image
# 2018-02   : Drop one array dimension, use numpy broadcast multiplication
# 2015-12-16: Modified to calculate the forward Abel transform
# 2015-12-03: Vectorization and code improvements Dan Hickstein and
#             Roman Yurchak
#             Previously the algorithm iterated over the rows of the image
#             now all of the rows are calculated simultaneously, which provides
#             the same result, but speeds up processing considerably.
#
# Historically, this algorithm was adapted by Jason Gascooke from ref. [2] in:
#
#  J. R. Gascooke PhD Thesis:
#   "Energy Transfer in Polyatomic-Rare Gas Collisions and Van Der Waals
#    Molecule Dissociation", Flinders University, 2000.
#
# Implemented in Python, with image quadrant co-adding, by Stephen Gibson (ANU)
# Significant code/speed improvements due to Dan Hickstein and Roman Yurchak
#
# Stephen Gibson - Australian National University, Australia
#
#############################################################################


def hansenlaw_transform(image, dr=1, direction='inverse', hold_order=0,
                        **kwargs):
    r"""Forward/Inverse Abel transformation using the algorithm from

    E. W. Hansen,
    "Fast Hankel transform algorithm",
    `IEEE Trans. Acoust. Speech Signal Proc. 33, 666–671 (1985)
    <https://dx.doi.org/10.1109/TASSP.1985.1164579>`__

    and

    E. W. Hansen, P.-L. Law,
    "Recursive methods for computing the Abel transform and its inverse",
    `J. Opt. Soc. Am. A 2, 510–520 (1985)
    <https://dx.doi.org/10.1364/JOSAA.2.000510>`__.

    This function performs the Hansen–Law transform on only one "right-side"
    image::

        Abeltrans = abel.hansenlaw.hansenlaw_transform(image, direction='inverse')

    .. note::  Image should be a right-side image, like this::

        .         +--------      +--------+
        .         |      *       | *      |
        .         |   *          |    *   |  <---------- im
        .         |  *           |     *  |
        .         +--------      o--------+
        .         |  *           |     *  |
        .         |   *          |    *   |
        .         |     *        | *      |
        .         +--------      +--------+

        In accordance with all PyAbel methods the image origin ``o`` is
        defined to be mid-pixel i.e. an odd number of columns, for the
        full image.


    For the full image transform, use the
    :class:`abel.Transform<abel.transform.Transform>`.

    Inverse Abel transform::

      iAbel = abel.Transform(image, method='hansenlaw').transform

    Forward Abel transform::

      fAbel = abel.Transform(image, direction='forward', method='hansenlaw').transform


    Parameters
    ----------
    image : 1D or 2D numpy array
        Right-side half-image (or quadrant). See figure below.

    dr : float
        Sampling size, used for Jacobian scaling.
        Default: `1` (appliable for pixel images).

    direction : string 'forward' or 'inverse'
        ``forward`` or ``inverse`` Abel transform.
        Default: 'inverse'.

    hold_order : int 0 or 1
        The order of the hold approximation, used to evaluate the state equation
        integral. 
        `0` assumes a constant intensity across a pixel (between grid points)
        for the driving function (the image gradient for the inverse transform,
        or the original image, for the forward transform).
        `1` assumes a linear intensity variation between grid points, which may
        yield a more accurate transform for some functions (see PR 211).
        Default: `0`.

    Returns
    -------
    aim : 1D or 2D numpy array
        forward/inverse Abel transform half-image


    """

    # state equation integral_r0^r (epsilon/r)^(lamda+a) d\epsilon
    def I(n, lam, a):
        integral = np.empty((n.size, lam.size))

        ratio = n/(n-1)
        if a == 0:
            integral[:, 0] = -np.log(ratio)  # special case, lam=0

        ra = (n-1)**a
        k0 = not a  # 0 or 1

        for k, lamk in enumerate((lam+a)[k0:], start=k0):
            integral[:, k] = ra*(1 - ratio**lamk)/lamk

        return integral

    # parameters for Abel transform system model, Table 1.
    h = np.array([0.318, 0.19, 0.35, 0.82, 1.8, 3.9, 8.3, 19.6, 48.3])
    lam = np.array([0.0, -2.1, -6.2, -22.4, -92.5, -414.5, -1889.4, -8990.9,
                    -47391.1])

    image = np.atleast_2d(image)   # 2D input image
    aim = np.empty_like(image)  # Abel transform array
    rows, cols = image.shape

    if direction == 'forward':
        # copy image for the driving function, include Jacobian factor,
        drive = -2*dr*np.pi*np.copy(image)
        a = 1  # integration increases lambda + 1
    else:  # inverse Abel transform
        if hold_order == 0:
            # better suits sharp structure - see issue #249
            drive = np.zeros_like(image)
            drive[:, :-1] = (image[:, 1:] - image[:, :-1])/dr
        else:
            # hold_order=1 prefers gradient
            drive = np.gradient(image, dr, axis=-1)
        a = 0  # due to 1/piR factor

    n = np.arange(cols-1, 1, -1)

    phi = np.empty((n.size, h.size))
    for k, lamk in enumerate(lam):
        phi[:, k] = (n/(n-1))**lamk

    gamma0 = I(n, lam, a)*h

    if hold_order == 0:  # Hansen (& Law) zero-order hold approximation
        B1 = gamma0
        B0 = gamma0*0  # empty array

    else:  # Hansen first-order hold approximation
        gamma1 = I(n, lam, a+1)*h

        B0 = gamma1 - gamma0*(n-1)[:, None]  # f_n
        B1 = gamma0*n[:, None] - gamma1  # f_n-1

    # Hansen Abel transform  --------------------
    x = np.zeros((h.size, rows))

    for indx, col in enumerate(n-1):
        x = phi[indx][:, None]*x + B0[indx][:, None]*drive[:, col+1]\
                                 + B1[indx][:, None]*drive[:, col]
        aim[:, col] = x.sum(axis=0)

    # missing column at each side of image
    aim[:, 0] = aim[:, 1]
    aim[:, -1] = aim[:, -2]

    if rows == 1:
        aim = aim[0]  # flatten to a vector

    return aim



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import itertools

import numpy as np
from numpy.testing import assert_equal, assert_allclose
from scipy.ndimage.interpolation import shift

import abel
from abel.tools.center import find_origin, center_image, set_center


def test_find_origin():
    """
    Test find_origin methods.
    """
    size = [12, 13]
    row, col = 5.4, 6.6  # origin
    w = 3.0  # gaussian width parameter (sqrt(2) * sigma)
    for rows in size:
        y2 = ((np.arange(rows) - row) / w)**2
        for cols in size:
            x2 = ((np.arange(cols) - col) / w)**2
            data = np.exp(-(x2 + y2[:, None]))
            axes = (1, 0)
            # (not testing trivial 'image_center', which does not find origin)
            for method in ['com', 'convolution', 'gaussian', 'slice']:
                origin = find_origin(data, method, axes)
                ref = (row if 0 in axes else rows // 2,
                       col if 1 in axes else cols // 2)
                tol = 0.2  # 'convolution' rounds to 0.5 pixels
                assert_allclose(origin, ref, atol=tol, verbose=False,
                                err_msg='-> {} x {}, method = {}, axes = {}: '
                                        'origin = {} not equal {}'.
                                        format(rows, cols, method, axes,
                                               origin, ref))


def test_set_center_int():
    """
    Test whole-pixel shifts.
    """
    # input sizes
    size = [4, 5]
    # input size, crop, origin -> output elements
    param = {4: {'maintain_size': [[None, '1234'],
                                   [0,    '0012'],
                                   [1,    '0123'],
                                   [2,    '1234'],
                                   [3,    '2340']],
                 'valid_region':  [[None, '1234'],
                                   [0,    '1'],
                                   [1,    '123'],
                                   [2,    '234'],
                                   [3,    '4']],
                 'maintain_data': [[None, '1234'],
                                   [0,    '0001234'],
                                   [1,    '01234'],
                                   [2,    '12340'],
                                   [3,    '1234000']]},
             5: {'maintain_size': [[None, '12345'],
                                   [0,    '00123'],
                                   [1,    '01234'],
                                   [2,    '12345'],
                                   [3,    '23450'],
                                   [4,    '34500']],
                 'valid_region':  [[None, '12345'],
                                   [0,    '1'],
                                   [1,    '123'],
                                   [2,    '12345'],
                                   [3,    '345'],
                                   [4,    '5']],
                 'maintain_data': [[None, '12345'],
                                   [0,    '000012345'],
                                   [1,    '0012345'],
                                   [2,    '12345'],
                                   [3,    '1234500'],
                                   [4,    '123450000']]}}
    # all size combinations
    for rows, cols in itertools.product(size, repeat=2):
        # test data: consecutive numbers from 1, row by row
        data = (np.arange(rows * cols) + 1).reshape((rows, cols))
        # all crop options
        for crop in ['maintain_size', 'valid_region', 'maintain_data']:
            # all origin rows
            for row, rref in param[rows][crop]:
                # vector or reference rows
                rref = np.array([int(n) for n in rref])
                # all origin columns
                for col, cref in param[cols][crop]:
                    # vector of reference columns
                    cref = np.array([int(n) for n in cref])
                    # reference array
                    ref = (rref[:, None] - 1) * cols + cref
                    ref[rref == 0] = 0
                    ref[:, cref == 0] = 0
                    # check set_center() result
                    result = set_center(data, (row, col), crop=crop)
                    assert_equal(result, ref, verbose=False,
                                 err_msg='-> {} x {}, origin = {}, crop = {}\n'
                                         'result =\n{}\n'
                                         'must be =\n{}'.
                                         format(rows, cols, (row, col), crop,
                                                result, ref))


def test_set_center_float():
    """
    Test fractional shifts.
    """
    # input sizes
    size = [10, 11]
    # default origin coordinate (substituting None)
    default = 5.0
    # input size, origin, crop -> output size, non-zero range
    param = {10: [(None, {'maintain_size': [10, (0, 10)],
                          'valid_region':  [10, (0, 10)],
                          'maintain_data': [10, (0, 10)]}),
                  (2.5,  {'maintain_size': [10, (2, 10)],
                          'valid_region':  [5,  (0,  5)],
                          'maintain_data': [15, (4, 15)]}),
                  (3.5,  {'maintain_size': [10, (1, 10)],
                          'valid_region':  [7,  (0,  7)],
                          'maintain_data': [13, (2, 13)]}),
                  (4.5,  {'maintain_size': [10, (0, 10)],
                          'valid_region':  [9,  (0,  9)],
                          'maintain_data': [11, (0, 11)]}),
                  (5.5,  {'maintain_size': [10, (0, 10)],
                          'valid_region':  [7,  (0,  7)],
                          'maintain_data': [13, (0, 11)]}),
                  (6.5,  {'maintain_size': [10, (0,  9)],
                          'valid_region':  [5,  (0,  5)],
                          'maintain_data': [15, (0, 11)]})],
             11: [(None, {'maintain_size': [11, (0, 11)],
                          'valid_region':  [11, (0, 11)],
                          'maintain_data': [11, (0, 11)]}),
                  (3.5,  {'maintain_size': [11, (1, 11)],
                          'valid_region':  [7,  (0,  7)],
                          'maintain_data': [15, (3, 15)]}),
                  (4.5,  {'maintain_size': [11, (0, 11)],
                          'valid_region':  [9,  (0,  9)],
                          'maintain_data': [13, (1, 13)]}),
                  (5.5,  {'maintain_size': [11, (0, 11)],
                          'valid_region':  [9,  (0,  9)],
                          'maintain_data': [13, (0, 12)]}),
                  (6.5,  {'maintain_size': [11, (0, 10)],
                          'valid_region':  [7,  (0,  7)],
                          'maintain_data': [15, (0, 12)]})]}
    w = 2.0  # gaussian width parameter (sqrt(2) * sigma)
    # all size combinations
    for rows, cols in itertools.product(size, repeat=2):
        # all origin "rows"
        for row, rparam in param[rows]:
            y2 = ((np.arange(rows) - (row or default)) / w)**2
            # all origin "columns"
            for col, cparam in param[cols]:
                x2 = ((np.arange(cols) - (col or default)) / w)**2
                # test data: gaussian centered at (row, col)
                data = np.exp(-(x2 + y2[:, None]))
                # all crop options
                for crop in ['maintain_size', 'valid_region', 'maintain_data']:
                    # check set_center() result
                    result = set_center(data, (row, col), crop=crop)
                    refrows, rrange = rparam[crop]
                    refcols, crange = cparam[crop]
                    refshape = (refrows, refcols)
                    refrange = (slice(*rrange), slice(*crange))
                    reforigin = (refrows // 2 if row else default,
                                 refcols // 2 if col else default)
                    msg = '-> {} x {}, origin = {}, crop = {}: '.\
                          format(rows, cols, (row, col), crop)
                    # shape
                    assert_equal(result.shape, refshape, verbose=False,
                                 err_msg=msg + 'shape {} not equal {}'.
                                               format(result.shape, refshape))
                    # non-zero data
                    assert_equal(result[refrange] != 0, True,
                                 err_msg=msg + 'zeros in non-zero range')
                    # zero padding
                    tmp = result.copy()
                    tmp[refrange] = 0
                    assert_equal(tmp, 0, err_msg=msg +
                                 'non-zeros outside non-zero range')
                    # gaussian center
                    origin = find_origin(result, 'gaussian')
                    assert_allclose(origin, reforigin, atol=0.01,
                                    verbose=False, err_msg=msg +
                                    'shifted center {} not equal {}'.
                                    format(origin, reforigin))


def test_set_center_axes():
    """
    Test "None" origin components and axes selection.
    """
    for N in [4, 5]:
        data = np.arange(N**2).reshape((N, N))
        c = N // 2
        msg = '-> N = {}, '.format(N)
        assert_equal(set_center(data, (None, None)),
                     data,
                     err_msg=msg + '(None, None)')
        assert_equal(set_center(data, (0, 0), axes=[]),
                     data,
                     err_msg=msg + '(0, 0), axes=[]')
        assert_equal(set_center(data, (0, None)),
                     set_center(data, (0, c)),
                     err_msg=msg + '(0, None)')
        assert_equal(set_center(data, (None, 0)),
                     set_center(data, (c, 0)),
                     err_msg=msg + '(None, 0)')
        assert_equal(set_center(data, (0, 0), axes=0),
                     set_center(data, (0, c)),
                     err_msg=msg + '(0, 0), axes=0')
        assert_equal(set_center(data, (0, 0), axes=1),
                     set_center(data, (c, 0)),
                     err_msg=msg + '(0, 0), axes=1')


def test_set_center_order():
    """
    Test rounding for order = 0 and exact output for order = 1.
    """
    data = data = np.ones((5, 5))
    origin = np.array([1.9, 2.2])
    # check origin rounding for order = 0
    assert_equal(set_center(data, origin, order=0),
                 set_center(data, origin.round()),
                 err_msg='-> order = 0 not equal round(origin)')
    # check output for order = 1:
    # maintain_size
    result = set_center(data, origin, 'maintain_size', order=1)
    ref = np.outer([0.9, 1, 1, 1, 1],
                   [1, 1, 1, 1, 0.8])
    assert_allclose(result, ref,
                    err_msg='-> crop = maintain_size, order = 1')
    # valid_region
    result = set_center(data, origin, 'valid_region', order=1)
    ref = np.ones((3, 3))
    assert_allclose(result, ref,
                    err_msg='-> crop = valid_region, order = 1')
    # maintain_data
    result = set_center(data, origin, 'maintain_data', order=1)
    ref = np.outer([0, 0.9, 1, 1, 1, 1, 0.1],
                   [0.2, 1, 1, 1, 1, 0.8, 0])
    assert_allclose(result, ref,
                    err_msg='-> crop = maintain_data, order = 1')


def test_center_image():

    # BASEX sample image, Gaussians at 10, 15, 20, 70,85, 100, 145, 150, 155
    # image width, height n = 361, origin = (180, 180)
    IM = abel.tools.analytical.SampleImage(n=361, name="dribinski").image

    # artificially displace origin, now at (179, 182)
    IMx = shift(IM, (-1, 2))
    true_origin = (179, 182)

    # find_origin using 'slice' method
    origin = find_origin(IMx, method="slice")

    assert_allclose(origin, true_origin, atol=1)

    # find_origin using 'com' method
    origin = find_origin(IMx, method="com")

    assert_allclose(origin, true_origin, atol=1)

    # check single axis - vertical
    # center shifted image IMx in the vertical direction only
    IMc = center_image(IMx, method="com", axes=1)
    # determine the origin
    origin = find_origin(IMc, method="com")

    assert_allclose(origin, (179, 180), atol=1)

    # check single axis - horizontal
    # center shifted image IMx in the horizontal direction only
    IMc = center_image(IMx, method="com", axes=0)
    origin = find_origin(IMc, method="com")

    assert_allclose(origin, (180, 182), atol=1)

    # check even image size returns odd
    # drop off one column, to make an even column image
    IM = IM[:, :-1]
    m, n = IM.shape

    IMy = center_image(IM, method="slice", odd_size=True)

    assert_allclose(IMy.shape, (m, n-1))


if __name__ == "__main__":
    test_find_origin()
    test_set_center_axes()
    test_set_center_int()
    test_set_center_float()
    test_set_center_order()
    test_center_image()



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import os.path
from glob import glob
from math import exp, log

import numpy as np
from scipy.special import gammaln
from scipy.linalg import inv

import abel
from abel.tools.polynomial import PiecewisePolynomial

#############################################################################
# This is adapted from the BASEX Matlab code provided by the Reisler group.
#
# Please cite: "Reconstruction of Abel-transformable images:
# The Gaussian basis-set expansion Abel transform method"
# V. Dribinski, A. Ossadtchi, V. A. Mandelshtam, and H. Reisler,
# Review of Scientific Instruments 73, 2634 (2002).
# https://doi.org/10.1063/1.1482156
#
#
# 2018-10-29
#   MR added forward transform.
# 2018-10-27
#   MR exposed BASIS_SET_CUTOFF for speed/accuracy control.
# 2018-10-07
#   MR added intensity correction.
#   Also smaller basis sets are now reused when generating larger ones.
#   Removed unnecessary scipy methods (scipy.dot is actually numpy.dot).
# 2018-10-03
#   MR completely rewrote basis generation (half-width, efficiency).
#   Switched from (n, nbf) to (n, sigma) basis specification;
#   nbf is now determined automatically, and nbf != n actually works.
#   The regularization parameter now differs from BASEX.exe by a factor
#   of 4/pi (BASEX.exe had an incorrect prefactor in basis projections).
# 2018-09-29
#   MR improved loading cached basis sets:
#   If the required basis is not available, but a larger compatible is,
#   then the latter will be loaded and cropped to the required size.
# 2018-09-19
#   MR switched to half-width transform.
#   The results now match BASEX.exe.
# 2018-09-08
#   MR enabled Tikhonov regularization.
#   Also, basis and transform matrices are now kept in memory
#   between invocations and reloaded/recalculated only when needed.
# Version 0.62 - 2016-03-07
#   DH changed all linear algebra steps from numpy to scipy
#   Scipy uses fortran fftpack, so it will be faster on some systems
#   or the same speed as numpy in most cases.
# Version 0.61 - 2016-02-17
#   Major reformatting - removed up/down symmetic version and
#   made basex work with only one quadrant.
# Version 0.6 - 2015-12-01
#   Another major code reformatting
# Version 0.5 - 2015-11-16
#   Code cleanup
# Version 0.4 - 2015-05-0
#   Major code update see pull request
#   https://github.com/DanHickstein/pyBASEX/pull/3
# Version 0.3 - 2015-02-01
#   Added documentation
# Version 0.2 - 2014-10-09
#   Adding a "center_and_transform" function to make things easier
# Versions 0.1 - 2012
#   First port to Python
#
#############################################################################


def basex_transform(data, sigma=1.0, reg=0.0, correction=True, basis_dir='',
                    dr=1.0, verbose=True, direction='inverse'):
    """
    This function performs the BASEX (BAsis Set EXpansion)
    Abel transform. It works on a "right side" image. I.e.,
    it works on just half of a cylindrically symmetric
    object, and ``data[0,0]`` should correspond to a central pixel.
    To perform a BASEX transform on a whole image, use ::

        abel.Transform(image, method='basex', direction='inverse').transform

    This BASEX implementation only works with images that have an
    odd-integer full width.

    Parameters
    ----------
    data : m × n numpy array
        the image to be transformed.
        ``data[:,0]`` should correspond to the central column of the image.
    sigma : float
        width parameter for basis functions, see equation (14) in the article.
        Determines the number of basis functions (**n**/**sigma** rounded).
        Can be any positive number, but using **sigma** < 1
        is not very meaningful and requires regularization.
    reg : float
        regularization parameter, square of the Tikhonov factor.

            ``reg=0`` means no regularization,

            ``reg=100`` is a reasonable value for megapixel images.

        Forward transform requires regularization only if **sigma** < 1,
        and **reg** should be ≪ 1.
    correction : boolean
        apply intensity correction in order to reduce method artifacts
        (intensity normalization and oscillations)
    basis_dir : str or None
        path to the directory for saving / loading the basis sets. Use ``''``
        for the default directory. If ``None``, the basis set will not be
        loaded from or saved to disk.
    dr : float
        size of one pixel in the radial direction.
        This only affects the absolute scaling of the transformed image.
    verbose : boolean
        determines whether statements should be printed
    direction : str: ``'forward'`` or ``'inverse'``
        type of Abel transform to be performed

    Returns
    -------
    recon : m × n numpy array
        the transformed (half) image
    """

    # make sure that the data is the right shape (1D must be converted to 2D):
    data = np.atleast_2d(data)
    h, w = data.shape

    if data.shape[0] == 1:
        data_ndim = 1
    elif data.shape[1] == 1:
        raise ValueError('Wrong input shape for '
                         'data {0}, should be (N1, N2) '
                         'or (1, N), not (N, 1)'.format(data.shape))
    else:
        data_ndim = 2

    n = data.shape[1]

    # load the basis sets and compute the transform matrix
    A = get_bs_cached(n, sigma=sigma, reg=reg, correction=correction,
                      basis_dir=basis_dir, dr=dr, verbose=verbose,
                      direction=direction)

    # do the actual transform
    recon = basex_core_transform(data, A)

    if data_ndim == 1:  # taking the middle row, since the rest are zeroes
        recon = recon[recon.shape[0] - recon.shape[0]//2 - 1]  # ??
    if h == 1:
        return recon
    else:
        return recon


def basex_core_transform(rawdata, A):
    """
    Internal function that does the actual BASEX transform.
    It requires that the transform matrix be passed.

    Parameters
    ----------
    rawdata : m × n numpy array
        right half (with the axis) of the input image.
    A : n × n numpy array
        2D array given by the transform-calculation function

    Returns
    -------
    IM : m × n numpy array
        the Abel-transformed image
    """

    # Note that our images are stored with matrix rows corresponding to
    # horizontal image lines. This is consistent with the Matlab and C++
    # BASEX implementations, but is the opposite to the article notation.
    # Thus all matrices are transposed and multiplied backwards with
    # respect to the article.
    # The vertical transform is not applied, since without regularization
    # its overall effect is an identity transform.

    # transform the image
    return rawdata.dot(A)


def _get_A(M, Mc, reg, direction):
    """ Internal helper function.
        Calculates the forward/inverse Abel-transform matrix
        from basis matrices for given regularization parameter.
    """
    # A   is the overall (horizontal) transform matrix,
    #     corresponds to A^T Z in the article.
    # reg corresponds to q_1^2 in the article.

    # basis matrices for input and output spaces
    if direction == 'forward':
        Bi, Bo = Mc, M  # image -> projection
    else:  # 'inverse'
        Bi, Bo = M, Mc  # projection -> image

    n, nbf = M.shape
    if reg == 0.0 and nbf == n:
        A = inv(Bi.T).dot(Bo.T)
    else:
        # square of Tikhonov matrix
        E = np.diag([reg] * nbf)
        # regularized inverse of input basis
        R = Bi.dot(inv((Bi.T).dot(Bi) + E))
        # {expansion coefficients} = input . R
        # output = {expansion coefficients} . {output basis}
        # so: output = input . (R . {output basis})
        #     output = input . A
        #     A = R . {output basis}
        # A is the matrix of the Abel transform
        A = R.dot(Bo.T)

    return A


def _nbf(n, sigma):
    """
    Internal helper function.
    Calculates the number of basis functions **nbf** from
    the half-image width **n** and the basis width parameter **sigma**.
    """
    return int(round(n / sigma))


# Cached matrices and their parameters
# basis set
_bs_prm = None   # [n, sigma]
_bs = None       # [M, Mc]
# forward transform
_trf_prm = None  # [reg, correction, dr]
_trf = None      # Af
# inverse transform
_tri_prm = None  # [reg, correction, dr]
_tri = None      # Ai

def get_bs_cached(n, sigma=1.0, reg=0.0, correction=True, basis_dir='', dr=1.0,
                  verbose=False, direction='inverse'):
    """
    Internal function.

    Gets BASEX basis sets, using the disk as a cache
    (i.e. load from disk if they exist,
    if not, calculate them and save a copy on disk)
    and calculates the transform matrix.
    To prevent saving the basis sets to disk, set ``basis_dir=None``.
    Loaded/calculated matrices are also cached in memory.

    Parameters
    ----------
    n : int
        Abel transform will be performed on an **n** pixels wide area
        of the (half) image
    sigma : float
        width parameter for basis functions
    reg : float
        regularization parameter
    correction : boolean
        apply intensity correction.
        Corrects wrong intensity normalization (seen for narrow basis sets),
        intensity oscillations (seen for broad basis sets),
        and intensity drop-off near *r* = 0 due to regularization.
    basis_dir : str or None
        path to the directory for saving / loading the basis sets. Use ``''``
        for the default directory. If ``None``, the basis sets will not be
        loaded from or saved to disk.
    dr : float
        pixel size. This only affects the absolute scaling of the output.
    verbose : boolean
        determines whether statements should be printed
    direction : str: ``'forward'`` or ``'inverse'``
        type of Abel transform to be performed

    Returns
    -------
    A : n × n numpy array
        matrix of the Abel transform (forward or inverse)
    """

    global _bs_prm, _bs, _trf_prm, _trf, _tri_prm, _tri

    sigma = float(sigma)  # (ensure FP format)
    nbf = _nbf(n, sigma)
    M = None
    # Check whether basis for these parameters is already loaded
    if _bs_prm == [n, sigma]:
        if verbose:
            print('Using memory-cached basis sets')
        M, Mc = _bs
    else:  # try to load basis
        if basis_dir == '':
            basis_dir = abel.transform.get_basis_dir(make=True)

        if basis_dir is not None:
            basis_file = 'basex_basis_{}_{}.npy'.format(n, sigma)

            def full_path(file_name):
                return os.path.join(basis_dir, file_name)

            # Try to find a suitable existing basis set
            if os.path.exists(full_path(basis_file)):
                # have exactly needed
                best_file = basis_file
            else:
                # Find the best (smallest among sufficient)
                # and the largest (to extend if not sufficient)
                best_file = None
                best_n = sys.maxsize
                largest_file = None
                largest_n = 0
                mask = os.path.join(basis_dir,
                                    'basex_basis_*_{}.npy'.format(sigma))
                for f in glob(mask):
                    # extract basis image size (sigma was fixed above)
                    f_n = int(f.split('_')[-2])
                    # must be large enough and smaller than previous best
                    if n <= f_n < best_n:
                        # remember as new best
                        best_file = f
                        best_n = f_n
                    # largest must be just larger than previous
                    if f_n > largest_n:
                        # remember as new largest
                        largest_file = f
                        largest_n = f_n

            # If found, try to use it
            if best_file:
                if verbose:
                    print('Loading basis sets...')
                    # saved as a .npy file
                try:
                    M, Mc = np.load(full_path(best_file))
                    # crop if loaded larger
                    if M.shape != (n, nbf):
                        M = M[:n, :nbf]
                        Mc = Mc[:n, :nbf]
                        if verbose:
                            print('(cropped from {})'.format(best_file))
                except ValueError:
                    print('Cached basis file incompatible.')

        if M is None:  # generate the basis set
            if verbose:
                print('A suitable basis set was not found.',
                      'A new basis set will be generated.',
                      'This may take a few minutes.', sep='\n')
                if basis_dir is not None:
                    print('But don\'t worry, '
                          'it will be saved to disk for future use.')

            # Try to extend the largest available
            try:
                oldM, oldMc = np.load(full_path(largest_file))
                if verbose:
                    print('(extending {})'.format(largest_file))
            except:
                oldM = None  # (old Mc is not needed)

            M, Mc = _bs_basex(n, sigma, oldM, verbose=verbose)

            if basis_dir is not None:
                np.save(full_path(basis_file), (M, Mc))
                if verbose:
                    print('Basis set saved for later use to')
                    print('  {}'.format(basis_file))

        _bs_prm = [n, sigma]
        _bs = [M, Mc]
        _trf_prm = None
        _tri_prm = None

    # Check whether transform matrices for these parameters
    # are already created
    if direction == 'forward' and _trf_prm == [reg, correction, dr]:
        A = _trf
    elif direction == 'inverse' and _tri_prm == [reg, correction, dr]:
        A = _tri
    else:  # recalculate
        if verbose:
            print('Updating regularization...')
        A = _get_A(*_bs, reg=reg, direction=direction)
        if correction:
            if verbose:
                print('Calculating correction...')
            cor = get_basex_correction(A, sigma, direction)
            A = np.multiply(A, cor)
        if direction == 'forward':
            _trf_prm = [reg, correction, dr]
            _trf = A
        else:  # 'inverse'
            _tri_prm = [reg, correction, dr]
            _tri = A
        # apply intensity scaling, if needed
        if dr != 1.0:
            if direction == 'forward':
                A *= dr
            else:  # 'inverse'
                A /= dr

    return A


def cache_cleanup(select='all'):
    """
    Utility function.

    Frees the memory caches created by ``get_bs_cached()``.
    This is usually pointless, but might be required after working
    with very large images, if more RAM is needed for further tasks.

    Parameters
    ----------
    select : str
        selects which caches to clean:

        ``all`` (default)
            everything, including basis;
        ``forward``
            forward transform;
        ``inverse``
            inverse transform.

    Returns
    -------
    None
    """
    global _bs_prm, _bs, _trf_prm, _trf, _tri_prm, _tri

    if select == 'all':
        _bs_prm = None
        _bs = None
    if select in ('all', 'forward'):
        _trf_prm = None
        _trf = None
    if select in ('all', 'inverse'):
        _tri_prm = None
        _tri = None


def basis_dir_cleanup(basis_dir=''):
    """
    Utility function.

    Deletes basis sets saved on disk.

    Parameters
    ----------
    basis_dir : str or None
        absolute or relative path to the directory with saved basis sets. Use
        ``''`` for the default directory. ``None`` does nothing.

    Returns
    -------
    None
    """
    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=False)

    if basis_dir is None:
        return

    files = glob(os.path.join(basis_dir, 'basex_basis_*.npy'))
    for fname in files:
        os.remove(fname)


def get_basex_correction(A, sigma, direction):
    """
    Internal function.

    The default BASEX basis and the way its projection is calculated
    leads to artifacts in the reconstructed distribution --
    incorrect overall intensity for **sigma** = 1,
    intensity oscillations for other **sigma** values,
    intensity fluctuations (and drop-off for **reg** > 0) near *r* = 0.
    This function generates the intensity correction profile
    from the BASEX result for a step function with a soft edge (to avoid
    ringing) aligned with the last basis function.

    Parameters
    ----------
    A : n × n numpy array
        matrix of the Abel transform
    sigma : float
        basis width parameter
    direction : str: ``'forward'`` or ``'inverse'``
        type of the Abel transform

    Returns
    -------
    cor : 1 × n numpy array
        intensity correction profile
    """
    n = A.shape[0]
    nbf = _nbf(n, sigma)

    # Generate soft step function and its projection
    r = np.arange(float(n))
    # edge center, aligned with the last basis function
    c = (nbf - 0.5) * sigma
    # soft-edge halfwidth
    w = sigma
    # soft step: stitched constant (shelf) and 2 parabolas (soft edge)
    step = PiecewisePolynomial(r, [(0, c - w, [1]),
                                   (c - w, c, [1, 0, -1/2], c - w, w),
                                   (c, c + w, [0, 0, 1/2], c + w, w)])
    # (this is more numerically stable at large r than cubic smoothstep)

    # get BASEX Abel transform of the step
    # and set correction profile = expected / BASEX result
    if direction == 'forward':
        tran = basex_core_transform(step.func, A)
        cor = step.abel / tran
    else:  # 'inverse'
        tran = basex_core_transform(step.abel, A)
        cor = step.func / tran

    return cor


# The analytical expresion for the k-th basis-function projection
# involves a sum of k^2 terms, most of which are very small.
# Setting BASIS_SET_CUTOFF = c truncates this sum to ±cu terms
# around the maximum (u = x / sigma).
# The computation time is roughly proportional to this parameter,
# while the accuracy (at least for n, k < 10000) is as follows:
#   cutoff   relative error
#     4        < 2e-4
#     5        < 2e-6
#     6        < 7e-9
#     7        < 1e-11
#     8        < 6e-15
#     9        < 1e-15
# The last one reaches the 64-bit floating-point precision,
# so going beyond that is useless.
# See https://github.com/PyAbel/PyAbel/issues/230
BASIS_SET_CUTOFF = 9  # numerically exact

def _bs_basex(n=251, sigma=1.0, oldM=None, verbose=True):
    """
    Generates horizontal basis sets for the BASEX method.

    Parameters
    ----------
    n : int
        horizontal dimensions of the half-width image in pixels.
        Must include the axial pixel.
        See https://github.com/PyAbel/PyAbel/issues/34
    sigma : float
        width parameter for basis functions
    oldM : numpy array
        projected basis matrix for the same **sigma** but a smaller image size.
        Can be supplied to avoid recalculating matrix elements
        that are already available.

    Returns
    -------
    M, Mc : n × nbf numpy array
        Mc
            is the reconstructed-image basis rho_k(r_i) (~Gaussians),
            corresponds to Z^T in the article.
        M
            is the projected basis chi_k(x_i),
            corresponds to X^T in the article.
    """

    sigma = float(sigma)  # (ensure FP type)
    nbf = _nbf(n, sigma)  # number of basis functions

    if verbose:
        print('Generating horizontal BASEX basis sets for '
              'n = {}, sigma = {} (nbf = {}):'.format(n, sigma, nbf))
        print('k = 0...', end='')
        sys.stdout.flush()

    # Precompute tables of ln Gamma(...) terms;
    # notice that index i corresponds to argument i + 1 (and i + 1/2).
    maxk2 = (nbf - 1)**2
    # for Gamma(k^2 + 1) and Gamma(l + 1)
    lngamma = gammaln(np.arange(maxk2 + 1) + 1)
    # for Gamma(k^2 - l + 1) - Gamma(k^2 - l + 1/2)
    Dlngamma = lngamma - gammaln(np.arange(maxk2 + 1) + 1/2)

    # reduced coordinates u = x/sigma (or r/sigma) and their squares
    U = np.arange(float(n)) / sigma
    U2 = U * U

    Mc = np.empty((n, nbf))
    M = np.empty((n, nbf))
    # (indexing is Mc[r, k], M[x, k])
    old_n, old_nbf = 0, 0
    # reuse old elements, if available
    if oldM is not None:
        old_n, old_nbf = oldM.shape
        M[:old_n, :old_nbf] = oldM / sigma  # (full M will be *= sigma later)

    # Cases k = 0 and x = 0 (r = 0) are special, since general expressions
    # are valid only if considered as limits; here they are computed
    # separately, using expressions that result from taking these limits.
    # In all cases the sigma factor in projections is applied afterwards.

    # rho_0(r) = exp(-u^2)
    Mc[:, 0] = np.exp(-U2)
    # chi_0(x) = sqrt(pi) sigma exp(-u^2) = Gamma(1/2) sigma exp(-u^2)
    M[:, 0] = np.exp(gammaln(1/2) - U2)

    for k in range(1, nbf):
        k2 = k * k
        # prefactor ln[(e/k^2)^(k^2)]
        ek = (1 - log(k2)) * k2

        # Basis function rho_k(r)
        Mc[0, k] = 0
        Mc[1:, k] = np.exp(ek + np.log(U[1:]) * 2 * k2 - U2[1:])

        # Projected basis function chi_k(x)
        # full range of l
        L = np.arange(k2 + 1)
        # all ln Gamma(...) terms
        G = lngamma[k2] - lngamma[L] - Dlngamma[k2 - L]
        # Calculate chi_k(x) at each x_i
        M[0, k] = exp(ek + G[0])  # (u^(2l) = 1 for l = 0, otherwise 0)
        for i, u2 in enumerate(U2[1:], 1):
            # skip what was already filled
            if i < old_n and k < old_nbf:
                continue
            u = U[i]
            if u > k + 8:  # beyond outer shoulder
                M[i, k] = 0.0
                continue
            # index of the largest component
            lmax = min(int(u2), k2)
            # halfwidth of the important range
            delta = int(BASIS_SET_CUTOFF * (u + 2))
            # summation limits: ±delta from the maximum, but within [0, k^2]
            minl = max(0, lmax - delta)
            maxl = min(lmax + delta, k2)
            # list of ln[u^(2l)]
            lnu2L = log(u2) * L[minl:maxl+1]
            # sum over the important range
            M[i, k] = np.exp(ek - u2 + G[minl:maxl+1] + lnu2L).sum()

        if verbose and k % 50 == 0:
            print('{}...'.format(k), end='')
            sys.stdout.flush()

    if verbose:
        print(k + 1)

    M *= sigma  # applying the sigma factor

    return M, Mc



# -*- coding: utf-8 -*-
from __future__ import division
import numpy as np

##############################################################################
#
# Abel analytical function transform pairs:  source <-> projection
#
# 04-Jan-2018 Dan Hickstein - code improvements
# 20-Dec-2017 Stephen Gibson - adapted code for PyAbel
# 20-Nov-2015 Dhrubajyoti Das - python gist
#             https://github.com/PyAbel/PyAbel/issues/19#issuecomment-158244527
#
# Note: preferable to call these functions via the class method:
#        func = abel.tools.analytical.TransformPair(n, profile=#)
#   see abel/tools/analytical.py for Class attributes
#
##############################################################################

__doc__ = """
Analytical function Abel-transform pairs

profiles 1--7:
    G. C.-Y. Chan, Gary M. Hieftje,
    "Estimation of confidence intervals for radial emissivity and optimization
    of data treatment techniques in Abel inversion",
    `Spectrochimica Acta B 61, 31–41 (2006)
    <https://doi.org/10.1016/j.sab.2005.11.009>`_, Table 1.

Note:
    the transform pair functions are more conveniently accessed through
    :class:`abel.tools.analytical.TransformPair`::

        func = abel.tools.analytical.TransformPair(n, profile=nprofile)

    which sets the radial range `r` and provides attributes
    ``.func`` (source), ``.abel`` (projection), ``.r`` (radial range),
    ``.dr`` (step), ``.label`` (the profile name)


Parameters
----------
r : floats or numpy 1D array of floats
    value or grid to evaluate the function pair: ``0 < r < 1``

Returns
-------
source, projection : tuple of 1D numpy arrays of shape `r`
    source function profile (inverse Abel transform of projection),
    projection functon profile (forward Abel transform of source)

"""


def a(n, r):
    r""" coefficient

        .. math:: a_n = \sqrt{n^2 - r^2}

    """

    return np.sqrt(n*n - r*r)


def profile1(r):
    r"""**profile1**:
    C. J. Cremers, R. C. Birkebak,
    "Application of the Abel Integral Equation to Spectrographic Data",
    `Appl. Opt. 5, 1057–1064 (1966)
    <https://doi.org/10.1364/AO.5.001057>`_, Eq. (13).

    .. math::

        \epsilon(r) &= 0.75 + 12r^2 - 32r^3  & 0 \le r \le 0.25

        \epsilon(r) &= \frac{16}{27}(1 + 6r - 15r^2 + 8r^3)
                    & 0.25 \lt r \le 1

        I(r) &= \frac{1}{108}(128a_1 + a_{0.25}) + \frac{2}{27}r^2
                  (283a_{0.25} - 112a_1) + {}

        & \quad \frac{8}{9}r^2\left[4(1 + r^2)\ln\frac{1 + a_1}{r} -
          (4 + 31r^2)\ln\frac{0.25 + a_{0.25}}{r}\right] &  0 \le r \le 0.25

        I(r) &= \frac{32}{27}\left[a_1 - 7a_1 r + 3r^2(1 + r^2)
                \ln\frac{1 + a_1}{r}\right]  & 0.25 \lt r \le 1

    ..
              source                projection
        ┼+1.3                  ┼+1.3               
        │                      o   o               
        │     x                │     o             
        │   x   x              │       o           
        │ x                    │                   
        x         x            │         o         
        │           x          │                   
        │                      │           o       
        │             x        │                   
        │                      │             o     
        ┼+0─────────────x──┼   ┼+0─────────────o──┼
        0          r      +1   0          r      +1

    .. plot::

        from tools.transform_pairs import plot
        plot(1)
    """

    if np.any(r <= 0) or np.any(r > 1):
        raise ValueError('r must be 0 < r <= 1')

    if not hasattr(r, '__len__'):
        r = np.asarray([r])

    # left side r <= 0.25
    rl = r[r <= 0.25]

    # source
    source_l = 3/4 + 12*rl**2 - 32*rl**3

    # projection
    a4l = a(0.25, rl)
    a1l = a(1, rl)
    rl2 = rl**2

    proj_l = (128*a1l + a4l)/108 + (283*a4l - 112*a1l)*rl2*2/27 +\
             (4*(1 + rl2)*np.log((1 + a1l)/rl) -
             (4 + 31*rl2)*np.log((0.25 + a4l)/rl))*rl2*8/9

    # right side r > 0.25
    rr = r[r > 0.25]

    # source
    source_r = (16/27)*(1 + 6*rr - 15*rr**2 + 8*rr**3)

    # projection
    a1r = a(1, rr)
    rr2 = rr**2

    proj_r = (a1r - 7*a1r*rr2 + 3*rr2*(1 + rr2)*np.log((1 + a1r)/rr))*32/27

    source = np.concatenate((source_l, source_r))
    projection = np.concatenate((proj_l, proj_r))

    return source, projection


def profile2(r):
    r"""**profile2**:
    C. J. Cremers, R. C. Birkebak,
    "Application of the Abel Integral Equation to Spectrographic Data",
    `Appl. Opt. 5, 1057–1064 (1966)
    <https://doi.org/10.1364/AO.5.001057>`_, Eq. (11).

    .. math::

        \epsilon(r) &= 1 - 3r^2 + 2r^3 & 0 \le r \le 1

        I(r) &= a_1\left(1 - \frac{5}{2}r^2\right) +
                \frac{3}{2}r^4\ln\frac{1 + a_1}{r} & 0 \le r \le 1

    ..
              source                projection
        ┼+1.1                  ┼+1.1               
        x x                    o o                 
        │   x                  │   o               
        │     x                │     o             
        │       x              │                   
        │                      │       o           
        │         x            │                   
        │                      │         o         
        │           x          │           o       
        │             x        │                   
        ┼+0─────────────x──┼   ┼+0───────────o────┼
        0          r      +1   0          r      +1

    .. plot::

        from tools.transform_pairs import plot
        plot(2)
    """

    if np.any(r < 0) or np.any(r > 1):
        raise ValueError('r must be 0 <= r <= 1')

    if not hasattr(r, '__len__'):
        r = np.asarray([r])

    source = 1 - 3*r*r + 2*r**3

    a1 = a(1, r)
    projection = a1*(1 - r**2*5/2) + r**4*np.log((1 + a1)/r)*3/2

    return source, projection


def profile3(r):
    r"""**profile3**:
    C. J. Cremers, R. C. Birkebak,
    "Application of the Abel Integral Equation to Spectrographic Data",
    `Appl. Opt. 5, 1057–1064 (1966)
    <https://doi.org/10.1364/AO.5.001057>`_, Eq. (12).

    .. math::

        \epsilon(r) &= 1 - 2r^2  & 0 \le r \le 0.5

        \epsilon(r) &= 2(1 - r^2)^2 & 0.5 \lt r \le 1

        I(r) &= \frac{4a_1}{3}(1 + 2r^2) - \frac{2 a_{0.5}}{3}(1 + 8r^2) -
                4r^2\ln\frac{1 - a_1}{0.5 + a_{0.5}} & 0 \le r \le 0.5

        I(r) &= \frac{4a_1}{3}(1 + 2r^2) - 4r^2\ln\frac{1 - a_1}{r} &
                0.5 \lt r \le 1

    ..
              source                projection
        ┼+1.1                  ┼+1.1               
        x x x                  o o                 
        │                      │   o               
        │     x                │     o             
        │       x              │                   
        │                      │       o           
        │         x            │                   
        │                      │         o         
        │           x          │                   
        │                      │           o       
        ┼+0───────────x────┼   ┼+0───────────o────┼
        0          r      +1   0          r      +1

    .. plot::

        from tools.transform_pairs import plot
        plot(3)
    """

    if np.any(r < 0) or np.any(r > 1):
        raise ValueError('r must be 0 <= r <= 1')

    if not hasattr(r, '__len__'):
        r = np.asarray([r])

    # left side r <= 0.5
    rl = r[r <= 0.5]

    source_l = 1 - 2*rl**2

    a5l = a(0.5, rl)
    a1l = a(1, rl)
    # power rm**2 typo in Cremers
    proj_l = (4/3)*a1l*(1 + 2*rl**2) - (2/3)*a5l*(1 + 8*rl**2) -\
             4*rl**2*np.log((1 + a1l)/(0.5 + a5l))

    # right side r > 0.5
    rr = r[r > 0.5]
    a1r = a(1, rr)

    source_r = 2*(1 - rr)**2
    proj_r = (4/3)*a1r*(1 + 2*rr**2) - 4*rr**2*np.log((1 + a1r)/rr)

    source = np.concatenate((source_l, source_r))
    projection = np.concatenate((proj_l, proj_r))

    return source, projection


def profile4(r):
    r"""**profile4**:
    R. Álvarez, A. Rodero, M. C. Quintero,
    "An Abel inversion method for radially resolved measurements in the axial
    injection torch",
    `Spectochim. Acta B 57, 1665–1680 (2002)
    <https://doi.org/10.1016/S0584-8547(02)00087-3>`_, Eq. (10).

    Note:
        Published projection has misprints
        (“19\ **3**\ .30083” instead of “19\ **6**\ .30083” in both cases).

    .. math::

        \epsilon(r) &= 0.1 + 5.51r^2 - 5.25r^3 & 0 \le r \le 0.7

        \epsilon(r) &= -40.74 + 155.56r - 188.89r^2 + 74.07r^3
                    & 0.7 \lt r \le1

        I(r) &= 22.68862a_{0.7} - 14.811667a_1 + {}

             &\quad (217.557a_{0.7} - 196.30083a_1)r^2 +
             155.56r^2\ln\frac{1 + a_1}{0.7 + a_{0.7}} + {}

             &\quad r^4\left(55.5525\ln\frac{1 + a_1}{r} - 59.49\ln\frac{0.7 +
             a_{0.7}}{r}\right) & 0 \le r \le 0.7

        I(r) &= -14.811667a_1 - 196.30083a_1 r^2 + {}

             &\quad r^2(155.56 + 55.5525r^2) \ln\frac{1 + a_1}{r}
             & 0.7 \lt r \le 1

    ..
              source                projection
        ┼+2.2                  ┼+2.2       o       
        │                      │         o   o     
        │                      │       o       o   
        │                      │     o             
        │                      │                   
        │                      │ o o             o 
        │           x x        o                   
        │         x     x      │                   
        │       x              │                   
        │     x                │                   
        ┼+0─x─────────────x┼   ┼+0────────────────┼
        0          r      +1   0          r      +1

    .. plot::

        from tools.transform_pairs import plot
        plot(4)
    """

    def source_left(x):
        """Profile4 source x <= 0.7.

        """

        return 0.1 + 5.51*x**2 - 5.25*x**3

    def source_right(x):
        """Profile4 source x > 0.7.

        """

        return -40.74 + 155.56*x - 188.89*x**2 + 74.07*x**3

    def proj_left(x):
        """Profile4 projection x < 0.7 of right function part.

        """

        a7 = a(0.7, x)
        a1 = a(1, x)
        return 22.68862*a7 - 14.811667*a1 + (217.557*a7 - 196.30083*a1)*x**2 +\
               +155.56*x**2*np.log((1 + a1)/(0.7 + a7)) +\
               x**4*(55.5525*np.log((1 + a1)/x) - 59.49*np.log((0.7 + a7)/x))

    def proj_right(x):
        """Profile4 projection x > 0.7.

        """

        a1 = a(1, x)
        return -14.811667*a1 - 196.30083*a1*x**2 +\
               x**2*(155.56 + 55.5525*x**2)*np.log((1 + a1)/x)

    if np.any(r <= 0) or np.any(r > 1):
        raise ValueError('r must be 0 < r <= 1')

    if not hasattr(r, '__len__'):
        r = np.asarray([r])

    # left side r <= 0.7 of source, projection profile
    rl = r[r <= 0.7]

    source_l = source_left(rl)
    proj_l = proj_left(rl)

    # right side r > 0.7 of source, projection profile
    rr = r[r > 0.7]

    source_r = source_right(rr)
    proj_r = proj_right(rr)

    source = np.concatenate((source_l, source_r))
    projection = np.concatenate((proj_l, proj_r))

    return source, projection


def profile5(r):
    r"""**profile5**:
    M. J. Buie, J. T. P. Pender, J. P. Holloway, T. Vincent, P. L. G. Ventzek,
    M. L. Brake,
    `J. Quant. Spectrosc. Radiat. Transf. 55, 231–243 (1996)
    <https://doi.org/10.1016/0022-4073(95)00149-2>`_, Table 1, № 1.

    .. math::

        \epsilon(r) &= 1 & 0 \le r \le 1

        I(r) &= 2a_1 & 0 \le r \le 1

    ..
              source                projection
        ┼+2.1                  ┼+2.1               
        │                      │     o o           
        │                      │         o o       
        │                      │             o     
        │                      │                   
        │                      │               o   
        x x x x x x x x x x    │                   
        │                      │                 o 
        │                      │                   
        │                      │                   
        ┼+0────────────────┼   ┼+0────────────────┼
        0          r      +1   0          r      +1

    .. plot::

        from tools.transform_pairs import plot
        plot(5)
    """

    if np.any(r < 0) or np.any(r > 1):
        raise ValueError('r must be 0 <= r <= 1')

    if not hasattr(r, '__len__'):
        r = np.asarray([r])

    source = np.ones_like(r)
    projection = 2*a(1, r)

    return source, projection


def profile6(r):
    r"""**profile6**:
    M. J. Buie, J. T. P. Pender, J. P. Holloway, T. Vincent, P. L. G. Ventzek,
    M. L. Brake,
    `J. Quant. Spectrosc. Radiat. Transf. 55, 231–243 (1996)
    <https://doi.org/10.1016/0022-4073(95)00149-2>`_, Table 1, № 7.

    .. math::

        \epsilon(r) &= (1 - r^2)^{-\frac{3}{2}} \exp\left[1.1^2\left(
                        1 - \frac{1}{1 - r^2}\right)\right] & 0 \le r \le 1

        I(r) &= \frac{\sqrt{\pi}}{1.1a_1} \exp\left[1.1^2\left(
                        1 - \frac{1}{1 - r^2}\right)\right] & 0 \le r \le 1

    ..
              source                projection
        ┼+1.8                  ┼+1.8               
        │                      o o o               
        │                      │     o o           
        │                      │         o         
        │                      │                   
        x x x x x x x          │           o       
        │             x        │                   
        │                      │             o     
        │               x      │                   
        │                      │               o   
        ┼+0────────────────┼   ┼+0────────────────┼
        0          r      +1   0          r      +1

    .. plot::

        from tools.transform_pairs import plot
        plot(6)
    """

    if np.any(r < 0) or np.any(r > 1):
        raise ValueError('r must be 0 <= r <= 1')

    if not hasattr(r, '__len__'):
        r = np.asarray([r])

    source = np.exp(1.1**2*(1 - 1/(1 - r**2)))/np.sqrt(1 - r**2)**3
    projection = np.exp(1.1**2*(1 - 1/(1 - r**2)))*np.sqrt(np.pi)/1.1/a(1, r)

    return source, projection


def profile7(r):
    r"""**profile7**:
    M. J. Buie, J. T. P. Pender, J. P. Holloway, T. Vincent, P. L. G. Ventzek,
    M. L. Brake,
    `J. Quant. Spectrosc. Radiat. Transf. 55, 231–243 (1996)
    <https://doi.org/10.1016/0022-4073(95)00149-2>`_, Table 1, № 9 (divided by
    2).

    .. math::

        \epsilon(r) &= \frac{1}{2}(1 + 10r^2 - 23r^4 + 12r^6) & 0 \le r \le 1

        I(r) &= \frac{8}{105}a_1(19 + 34r^2 - 125r^4 + 72r^6) & 0 \le r \le 1

    ..
              source                projection
        ┼+1.7                  ┼+1.7               
        │                      o o o o o           
        │                      │         o         
        │                      │                   
        │       x x x          │           o       
        │     x       x        │                   
        │                      │             o     
        │   x                  │                   
        x x             x      │                   
        │                      │               o   
        ┼+0───────────────x┼   ┼+0────────────────┼
        0          r      +1   0          r      +1

    .. plot::

        from tools.transform_pairs import plot
        plot(7)
    """

    if np.any(r < 0) or np.any(r > 1):
        raise ValueError('r must be 0 <= r <= 1')

    if not hasattr(r, '__len__'):
        r = np.asarray([r])

    source = (1 + 10*r**2 - 23*r**4 + 12*r**6)/2
    projection = a(1, r)*(19 + 34*r**2 - 125*r**4 + 72*r**6)*8/105

    return source, projection



from __future__ import absolute_import
from __future__ import division

import numpy as np
from numpy.testing import assert_allclose, assert_equal

# to suppress deprecation warnings
from warnings import catch_warnings, simplefilter

from abel.tools.analytical import SampleImage
import abel.tools.vmi as vmi

# The Distributions class and related functions are tested separately in
# test_tools_distributions.py


def make_images(R):
    """
    Make test images (square, centered): 1, cos^2, sin^2
    """
    n = 2 * R + 1  # image size
    ones = np.ones((n, n))

    x = np.arange(float(n)) - R
    x2 = x**2
    r2 = x2 + x2[:, None]
    r2[R, R] = np.inf  # for sin = 0 at r = 0

    sin2 = x2 / r2
    cos2 = 1 - sin2

    return n, ones, cos2, sin2


def test_radial_intensity():
    """
    Basic tests of radial-intensity calculations.
    """
    func = {
        'int2D': vmi.angular_integration_2D,
        'int3D': vmi.angular_integration_3D,
        'avg2D': vmi.average_radial_intensity_2D,
        'avg3D': vmi.average_radial_intensity_3D
    }

    R = 50  # image radius
    n, ones, cos2, sin2 = make_images(R)

    def check(name, ref, rtol, kind, IM, **kwargs):
        r, intensity = vmi.radial_intensity(kind, IM, **kwargs)
        r_, intensity_ = func[kind](IM, **kwargs)
        assert_equal([r_, intensity_], [r, intensity],
                     err_msg='{}(...) != radial_intensity({}, ...)'.
                             format(func[kind], kind))
        # (ignoring pixels at r = 0 and 1, which can be poor)
        assert_allclose(intensity[2:R], ref(r[2:R]), rtol=rtol,
                        err_msg='-> {}, {}, {}'.format(kind, name, kwargs))

    def int2D(r):
        return 2 * np.pi * r

    check('ones', int2D, 0.01, 'int2D', ones)
    check('ones', int2D, 0.01, 'int2D', ones, dr=0.5)
    check('ones', int2D, 0.01, 'int2D', ones, dt=0.01)
    check('cos2', int2D, 0.01, 'int2D', cos2 * 2)
    check('sin2', int2D, 0.001, 'int2D', sin2 * 2)

    def int3D(r):
        return 4 * np.pi * r**2

    check('ones', int3D, 1e-7, 'int3D', ones)
    check('ones', int3D, 1e-7, 'int3D', ones, dr=0.5)
    check('ones', int3D, 1e-4, 'int3D', ones, dt=0.01)
    check('cos2', int3D, 0.01, 'int3D', cos2 * 3)
    check('sin2', int3D, 0.01, 'int3D', sin2 * 3/2)

    def avg(r):
        return np.ones_like(r)

    check('ones', avg, 0.01, 'avg2D', ones)
    check('ones', avg, 0.01, 'avg2D', ones, dr=0.5)
    check('ones', avg, 0.01, 'avg2D', ones, dt=0.01)
    check('cos2', avg, 0.01, 'avg2D', cos2 * 2)
    check('sin2', avg, 0.001, 'avg2D', sin2 * 2)

    check('ones', avg, 1e-7, 'avg3D', ones)
    check('ones', avg, 1e-7, 'avg3D', ones, dr=0.5)
    check('ones', avg, 1e-4, 'avg3D', ones, dt=0.01)
    check('cos2', avg, 0.01, 'avg3D', cos2 * 3)
    check('sin2', avg, 0.01, 'avg3D', sin2 * 3/2)


def test_angular_integration():
    """
    Basic tests of angular integration
    (in its current form, with wrong output).
    """
    R = 50  # image radius
    n, ones, cos2, sin2 = make_images(R)

    def check(name, ref, rtol, IM, **kwargs):
        with catch_warnings():
            simplefilter('ignore', category=DeprecationWarning)
            r, speeds = vmi.angular_integration(IM, **kwargs)
        # (ignoring pixels at r = 0 and 1, which can be poor)
        assert_allclose(speeds[2:R], ref(r[2:R]), rtol=rtol,
                        err_msg='-> {}, {}'.format(name, kwargs))

    # for Jacobian=True
    def spher(r):
        return 4 * r**2  # (in fact, must be 4 * np.pi * r**2)

    check('ones', spher, 0.001, ones)
    check('ones', spher, 0.001, ones, dr=0.5)
    check('ones', spher, 0.0001, ones, dt=0.01)
    check('cos2', spher, 0.01, cos2 * 3)
    check('sin2', spher, 0.01, sin2 * 3/2)

    # for Jacobian=False
    def polar(r):
        return np.full_like(r, 2 * np.pi)  # (in fact, should be 2 * np.pi * r)

    check('ones', polar, 0.02, ones, Jacobian=False)
    check('ones', polar, 0.02, ones, Jacobian=False, dr=0.5)
    check('ones', polar, 0.01, ones, Jacobian=False, dt=0.01)
    check('cos2', polar, 0.03, cos2 * 2, Jacobian=False)
    check('sin2', polar, 0.001, sin2 * 2, Jacobian=False)


def test_average_radial_intensity():
    """
    Basic tests of angular averaging.
    """
    R = 50  # image radius
    n, ones, cos2, sin2 = make_images(R)

    def check(name, ref, atol, IM, **kwargs):
        with catch_warnings():
            simplefilter('ignore', category=DeprecationWarning)
            r, intensity = vmi.average_radial_intensity(IM, **kwargs)
        assert_allclose(intensity[2:R], ref, atol=atol,
                        err_msg='-> {}, {}'.format(name, kwargs))

    check('ones', 1, 0.02, ones)
    check('ones', 1, 0.02, ones, dr=0.5)
    check('ones', 1, 0.01, ones, dt=0.01)
    check('cos2', 1/2, 0.02, cos2)
    check('sin2', 1/2, 0.001, sin2)


def test_anisotropy_parameter():
    """
    Basic tests of anisotropy fitting.
    """
    n = 100
    theta = np.linspace(-np.pi, np.pi, n, endpoint=False)

    ones = np.ones_like(theta)
    cos2 = np.cos(theta)**2
    sin2 = np.sin(theta)**2

    def check(name, ref, theta, intensity):
        beta, amplitude = vmi.anisotropy_parameter(theta, intensity)
        assert_allclose((beta[0], amplitude[0]), ref, atol=1e-8,
                        err_msg='-> ' + name)

    check('ones', (0, 1), theta, ones)
    check('cos2', (2, 1/3), theta, cos2)
    check('sin2', (-1, 2/3), theta, sin2)
    check('cos2sin2', (0, 1/8), theta, cos2 * sin2)


def test_radial_integration():
    """
    Basic test of radial integration.
    """
    # test image (not projected)
    IM = SampleImage(name='dribinski').image

    Beta, Amplitude, Rmidpt, _, _ = \
        vmi.radial_integration(IM, radial_ranges=([(65, 75), (80, 90), (95, 105)]))

    assert_equal(Rmidpt, [70, 85, 100], err_msg='Rmidpt')
    assert_allclose([b[0] for b in Beta], [0, 1.58, -0.85], atol=0.01,
                    err_msg='Beta')
    assert_allclose([a[0] for a in Amplitude], [880, 600, 560], rtol=0.01,
                    err_msg='Amplitude')


def test_toPES():
    """
    Basic test of toPES conversion.
    """
    # test image (not projected)
    IM = SampleImage(name='Ominus').image

    eBE, PES = vmi.toPES(*vmi.angular_integration_3D(IM),
                         energy_cal_factor=1.2e-5,
                         photon_energy=1.0e7/808.6, Vrep=-100,
                         zoom=IM.shape[-1]/2048)

    assert_allclose(eBE[PES.argmax()], 11780, rtol=0.001,
                    err_msg='-> eBE @ max PES')
    assert_allclose(PES.max(), 16570, rtol=0.001,
                    err_msg='-> max PES')


if __name__ == "__main__":
    test_radial_intensity()
    test_angular_integration()
    test_average_radial_intensity()
    test_anisotropy_parameter()
    test_radial_integration()
    test_toPES()
